'''
.. codeauthor:: Richard Plevin

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
from ..error import PygcamException, CommandlineError
from ..subcommand import SubcommandABC

class ConfigCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''List the values of configuration variables from
                  ~/.pygcam.cfg configuration file.'''}

        super(ConfigCommand, self).__init__('config', subparsers, kwargs, group='utils')

    def addArgs(self, parser):
        parser.add_argument('-d', '--useDefault', action='store_true',
                            help='Indicates to operate on the DEFAULT section rather '
                                 'than the project section.')

        parser.add_argument('-e', '--edit', action='store_true',
                            help='Edit the configuration file. The command given by the '
                            'value of config variable GCAM.TextEditor is run with the '
                            '.pygcam.cfg file as an argument.')

        parser.add_argument('name', nargs='?', default='',
                            help='Show the names and values of all parameters whose '
                            'name contains the given value. The match is case-insensitive. '
                            'If not specified, all variable values are shown.')

        parser.add_argument('-x', '--exact', action='store_true',
                            help='Treat the text not as a substring to match, but '
                            'as the name of a specific variable. Match is case-sensitive. '
                            'Prints only the value.')

        parser.add_argument('-t', '--test', action='store_true',
                            help='Test the settings in the configuration file to ensure '
                            'that the basic setup is ok, i.e., required parameters have '
                            'values that make sense. If specified, no variables are displayed.')

        return parser

    def testConfig(self, section):
        import os
        from ..config import getParam

        requiredDirs = ['SandboxRoot', 'ProjectRoot', 'ProjectDir',
                        'QueryDir', 'MI.Dir', 'RefWorkspace', 'TempDir']
        requiredFiles = ['ProjectXmlFile', 'RefConfigFile', 'MI.JarFile']
        optionalDirs  = ['UserTempDir']
        optionalFiles = ['RegionMapFile', 'RewriteSetsFile']

        dirVars  = requiredDirs  + optionalDirs
        fileVars = requiredFiles + optionalFiles

        optionalVars = optionalDirs + optionalFiles

        errors = []

        for item in dirVars + fileVars:
            var = 'GCAM.' + item
            value = getParam(var)

            if not value:
                if item in optionalVars:
                    continue
                print("Config variable", var, "is empty")

            elif not os.path.lexists(value) and item not in optionalVars:
                errors.append("%s = '%s' : non-existent path" % (var, value))

            elif not os.path.isfile(value) and item in fileVars:
                errors.append("%s = '%s' : does not refer to a file" % (var, value))

            elif not os.path.isdir(value) and item in dirVars:
                errors.append("%s = '%s' : does not refer to a directory" % (var, value))

            print('OK: %s = %s' %(var, value))
        if errors:
            print('')
            for error in errors:
                print('Error:', error)

    def run(self, args, tool):
        import re
        import subprocess
        from ..config import getParam, _ConfigParser, USR_CONFIG_FILE

        if args.edit:
            cmd = '%s %s/%s' % (getParam('GCAM.TextEditor'), getParam('Home'), USR_CONFIG_FILE)
            print(cmd)
            exitStatus = subprocess.call(cmd, shell=True)
            if exitStatus != 0:
                raise PygcamException("TextEditor command '%s' exited with status %s\n" % (cmd, exitStatus))
            return

        section = 'DEFAULT' if args.useDefault else args.projectName

        if not section:
            raise CommandlineError("Project was not specifed and GCAM.DefaultProject is not set")

        if section != 'DEFAULT' and not _ConfigParser.has_section(section):
            raise CommandlineError("Unknown configuration file section '%s'" % section)

        if args.test:
            self.testConfig(section)
            return

        if args.name and args.exact:
            value = getParam(args.name, section=section, raiseError=False)
            if value is not None:
                print(value)
            return

        # if no name is given, the pattern matches all variables
        pattern = re.compile('.*' + args.name + '.*', re.IGNORECASE)

        print("[%s]" % section)
        for name, value in sorted(_ConfigParser.items(section)):
            if pattern.match(name):
                # getParam does path translation for docker, if required
                value = getParam(name, section=section, raiseError=False)
                print("%25s = %s" % (name, value))


PluginClass = ConfigCommand



#!/usr/bin/env python
"""
.. "new" sub-command (creates a new project)

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016  Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
"""
from datetime import datetime

from ..subcommand import SubcommandABC

def driver(args, tool):
    import os
    import re
    import shutil
    from ..config import getParam, USR_CONFIG_FILE
    from ..utils import mkdirs, pathjoin, copyResource, getResource
    from ..error import CommandlineError
    from ..log import getLogger

    _logger = getLogger(__name__)

    projectName = args.name
    projectRoot = os.path.abspath(args.root) if args.root else getParam('GCAM.ProjectRoot')
    projectDir  = pathjoin(projectRoot, projectName)
    overwrite   = args.overwrite

    try:
        os.chdir(projectRoot)
    except Exception as e:
        raise CommandlineError("Can't change dir to '%s': %s" % (projectRoot, e))

    try:
        mkdirs(projectDir)
    except OSError as e:
        if e.errno == 17:   # already exists; ignore
            pass
    except Exception as e:
        raise CommandlineError("Can't create to '%s': %s" % (projectDir, e))

    try:
        os.chdir(projectDir)
    except Exception as e:
        raise CommandlineError("Can't chdir to '%s': %s" % (projectDir, e))

    dirsToCreate  = ['etc', 'queries', 'plugins', 'xmlsrc/baseline/xml']
    initsToCreate = ['__init__.py', 'xmlsrc/__init__.py']

    for name in dirsToCreate:
        _logger.debug('Creating %s/%s/', projectDir, name)
        mkdirs(name)

    for name in initsToCreate:  # create empty files
        _logger.debug('Creating %s/%s', projectDir, name)
        open(name, 'a').close() # append just in case it exists already

    etcDir = 'etc'
    exampleDir = pathjoin(etcDir, 'examples')
    projectEtc = pathjoin(projectDir, 'etc')

    # Provide example XML files and instructions by extracting these
    # as resources from the pygcam package.
    filesToCopy = ['project.xml', 'protection.xml',  'rewriteSets.xml', 'scenarios.xml',
                   'scenarios-iterator.xml', 'project2.xml', 'Instructions.txt',
                   'project-schema.xsd', 'protection-schema.xsd', 'queries-schema.xsd',
                   'rewriteSets-schema.xsd', 'scenarios-schema.xsd']

    for filename in filesToCopy:
        srcDir = etcDir if filename.endswith('.xsd') else exampleDir
        src = pathjoin(srcDir, filename)
        dst = pathjoin(projectEtc, filename)
        _logger.debug('Creating %s', dst)

        if filename == 'project.xml':
            # For project.xml, we change the project name to the name given by user
            if not overwrite and os.path.lexists(dst):
                raise CommandlineError("Refusing to overwrite '%s'" % os.path.abspath(dst))

            oldText = getResource(src)
            newText = re.sub('<project name="(\w*)">', '<project name="%s">' % projectName, oldText)
            with open(dst, 'w') as fp:
                fp.write(newText)
        else:
            copyResource(src, dst, overwrite=overwrite)

    if args.addToConfig:
        _logger.debug('Adding [%s] to %s', projectName, USR_CONFIG_FILE)

        cfgFile = pathjoin(getParam('Home'), USR_CONFIG_FILE)

        # Create backup copy of config file
        shutil.copyfile(cfgFile, cfgFile + '~')

        # If projectRoot defaulted to the value of GCAM.ProjectRoot, reference
        # the variable; otherwise use the pathname verbatim.
        dirName = projectDir if args.root else '%(GCAM.ProjectRoot)s/' + projectName

        # Add a project section to the .pygcam.cfg file
        with open(cfgFile, 'a') as f:
            f.write('\n[%s]\n' % projectName)
            f.write('# Added by "new" sub-command %s\n' % datetime.now().ctime())
            f.write('GCAM.ProjectDir = %s\n' % dirName)
            f.write('GCAM.ScenarioSetupFile = %(GCAM.ProjectDir)s/etc/scenarios.xml\n')
            f.write('GCAM.RewriteSetsFile   = %(GCAM.ProjectDir)s/etc/rewriteSets.xml\n')


class NewProjectCommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : '''Create the structure and files required for a new pygcam project.'''}
        super(NewProjectCommand, self).__init__('new', subparsers, kwargs, group='project')

    def addArgs(self, parser):
        # Positional args
        parser.add_argument('name',
                            help='''Create the structure for the named project, and copy example
                            XML files into the "etc" directory.''')

        parser.add_argument('-c', '--addToConfig', action='store_true',
                            help='''Add a section for the new project to $HOME/.pygcam.cfg after
                            making a backup of the file in $HOME/.pygcam.cfg~''')

        parser.add_argument('--overwrite', action='store_true',
                            help='''If files that are to be copied to the project directory exist, overwrite them.
                            By default, existing files are not overwritten.''')

        parser.add_argument('-r', '--projectRoot', dest='root', metavar='PATH',
                            help='''The directory in which to create a subdirectory for the named
                            project. Default is the value of config variable GCAM.ProjectRoot''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



Skin1 = 1
Skin2 = 2

Skin = Skin1

#
# Fonts
#
LatoFont  = 'Lato,arial,sans-serif'
DosisFont = 'Dosis,arial,sans-serif'
LabelFont = LatoFont

FontDict = {
    'PlotText' : LatoFont,
    'LabelFont': LatoFont,
    'HelpText' : DosisFont,
}

def getFont(name):
    return FontDict[name]

#
# Colors
#
_pageBgColor         = '#EEEEEE'
_plotBgColor         = '#191A1A'
_plotPaperBgColor    = '#020202'
_plotTextColor       = '#CCCCCC'
_widgetBgColor       = '#c5e8ed'

_widgetBgLight       = '#DAEAE2'
_pageBgLight         = 'whitesmoke'
_plotBgLight         = 'whitesmoke'
_plotPaperBgLight    = _widgetBgLight
_plotTextLight       = '#020202'


_helpTextColor       = '#888888'

_distBarsActiveColor   = 'rgb(192, 255, 245)'
_distBarsInactiveColor = 'rgba(192, 255, 245, 0.2)'

_distBarsActiveLight   = 'seagreen' # RGB(46, 139, 87)
_distBarsInactiveLight = 'rgba(46, 139, 87, 0.3)'

#
# Colors
#
ColorDict = {
    'PageBg' : _pageBgColor,
    'ChooserBgColor': _widgetBgColor,
    'DistBarsActive': _distBarsActiveColor,
    'DistBarsInactive' : _distBarsInactiveColor,
    'PaperBg' : _plotPaperBgColor,
    'PlotBg' : _plotBgColor,
    'PlotText' : _plotTextColor,
    'ResetButton' : _distBarsActiveColor,
    'ResetButtonText' : _plotPaperBgColor,
    'TornadoPlotBar' : 'goldenrod',
    'TornadoPlotLine' : _plotTextColor,
    'HelpText' : _helpTextColor,
    'MeanLineColor': 'red',
    'MedianLineColor': 'orange',
    'KDE': 'gold',
    'ScatterPoints': _distBarsActiveColor,
}

ColorDict2 = {
    'PageBg' : _pageBgLight,
    'ChooserBgColor': _widgetBgLight,
    'DistBarsActive': _distBarsActiveLight,
    'DistBarsInactive' : _distBarsInactiveLight,
    'PaperBg' : _widgetBgLight,
    'PlotBg' : _plotBgLight,
    'PlotText' : _plotTextLight,
    'ResetButton' : _distBarsActiveLight,
    'ResetButtonText' : 'white', #_plotPaperBgLight,
    'TornadoPlotBar' : _distBarsActiveLight,
    'TornadoPlotLine' : _plotTextLight,
    'HelpText' : _helpTextColor,
    'MeanLineColor': 'darkred',
    'MedianLineColor': 'navy',
    'KDE': 'goldenrod',
    'ScatterPoints': _distBarsActiveLight,
}

PlotMargin = dict(l=35, r=35, t=70, b=30, pad=4)
PlotMarginWithXTitle = dict(l=35, r=35, t=70, b=40, pad=4)

def getColor(name):
    d = ColorDict if Skin == Skin2 else ColorDict2
    return d[name]

#
# Styles
#
StyleDict = {
    'Page' : {
        'text-align': 'center',
        'background-color': getColor('PageBg'),
    },
    'Label' : {
        'font-family': LabelFont,
        'font-size'  : '14px',
        'font-weight': 'bold',
        'margin-top' : 20,
    },
    'HelpText' : {
        'font-family': getFont('HelpText'),
        'font-size': 11,
        'font-style': 'italic',
        'color': getColor('HelpText'),
    },
    'Chooser' : {
        'margin-right': 'auto',
        'margin-left' : 'auto',
    },
    'AutoMargins': {
        'margin-right': 'auto',
        'margin-left': 'auto',
    },
    'Button' : {
        'color'     : getColor('ResetButtonText'),
        'font-size' : 10,
        'box-shadow': '0 6px 10px 0 rgba(0, 0, 0, 0.24), 0 17px 50px 0 rgba(0, 0, 0, 0.19)',
        'border'    : 'none',
        'padding'   : '2px 4px',
        'text-align': 'center',
        'background-color': getColor('ResetButton'),
        'text-decoration' : None,
    },
    'PlotMargin': PlotMargin,
    'PlotMarginWithXTitle' : PlotMarginWithXTitle,

    'Plot' : dict(
        autosize=True,
        height=450,
        width=550,
        font=dict(family=getFont('PlotText'),
                  color=getColor('PlotText')),
        titlefont=dict(color=getColor('PlotText'),
                       size='16'),
        margin=PlotMargin,
        hovermode="closest",
        plot_bgcolor=getColor('PlotBg'),
        paper_bgcolor=getColor('PaperBg'),
        legend=dict(font=dict(size=10), orientation='h', bgcolor=getColor('PlotBg')),
        title='',
    ),
}

def getStyle(name):
    return StyleDict[name]

def updateStyle(style, **kwargs):
    """Update a copy of a dict or the dict for the given style"""
    if not isinstance(style, dict):
        style = getStyle(style)    # look up the style by name

    style = style.copy()
    style.update(**kwargs)
    return style



from lxml import etree as ET

from .error import FileFormatError
from .landProtection import Scenario
from .log import getLogger

_logger = getLogger(__name__)

PROTECTED = 'Protected'

def pp(elt):
    print(ET.tostring(elt, pretty_print=True))

def eltname(elt):
    return elt.get('name')


def _parse_land_basin(name):
    if name.startswith(PROTECTED):
        protected = PROTECTED
        name = name[len(PROTECTED):]
    else:
        protected = ''

    pos = name.rfind('_')
    landtype = name[:pos]
    basin = name[pos + 1:]
    return (landtype, basin, protected)


def _compose_land_basin(landtype, basin, protection):
    return "{}{}_{}".format(protection, landtype, basin)

def _set_land_values(land_leaf, vals):
    """
    Update allocation and landAllocation nodes with the values
    computed for each year.
    """
    nodes = land_leaf.xpath('.//allocation|.//landAllocation')
    for node in nodes:
        year = node.get('year')
        node.text = str(vals[year])

def _get_allocation(reg_dict, landtype, basin, protected=''):
    import pandas as pd

    land_key = protected + landtype + '_' + basin
    land_leaf = reg_dict[land_key]
    nodes = land_leaf.xpath('.//allocation[@year<1975]|.//landAllocation')
    [(node.get('year'), float(node.text)) for node in nodes]
    values = [(node.get('year'), float(node.text)) for node in nodes]
    ind, val = zip(*values)
    s = pd.Series(val, index=ind)
    return s

def _get_total_area(reg_dict, landtype, basin):
    unprot = _get_allocation(reg_dict, landtype, basin, protected='')
    prot   = _get_allocation(reg_dict, landtype, basin, protected=PROTECTED)
    return prot + unprot

def _landtype_basin_pairs(reg_dict):
    """Return the landtype and basin name, ignoring the protected field"""
    pairs = [_parse_land_basin(key)[0:2] for key in reg_dict.keys() if key.startswith(PROTECTED)]
    return pairs

def _get_land_leafs(tree, region):
    return tree.xpath('//region[@name="{}"]//UnmanagedLandLeaf'.format(region))

def _cache_land_nodes(tree, regions):
    d = {}
    for reg in regions:
        nodes = _get_land_leafs(tree, reg)
        d[reg] = {eltname(node) : node for node in nodes}
    return d

def _update_protection(reg_dict, landtype, basin, prot_vals, unprot_vals):
    def _upd(protected, vals):
        leaf_name = _compose_land_basin(landtype, basin, protected)
        land_leaf = reg_dict[leaf_name]
        _set_land_values(land_leaf, vals)

    _upd(PROTECTED, prot_vals)
    _upd('', unprot_vals)

def _protect_land(tree, prot_dict):
    node_dict = _cache_land_nodes(tree, prot_dict.keys())
    for (reg, prot_pairs) in prot_dict.iteritems():
        reg_dict = node_dict[reg]
        land_basin_pairs = _landtype_basin_pairs(reg_dict)

        for (landtype, prot_frac) in prot_pairs:
            for (ltype, basin) in land_basin_pairs:
                if landtype == ltype:
                    # print("Processing {}, {}, {}".format(reg, landtype, basin))
                    total = _get_total_area(reg_dict, landtype, basin)
                    prot_vals   = total * prot_frac
                    unprot_vals = total - prot_vals
                    _update_protection(reg_dict, landtype, basin, prot_vals, unprot_vals)

#
# Modified from landProtection.py method of same name
#
def protectLandTree(tree, scenarioName):
    """
    Apply the protection scenario `scenarioName` to the parsed XML file `tree`.
    This interface is provided so WriteFuncs (which are passed an open XMLInputFile)
    can apply protection scenarios.

    :param tree: (lxml ElementTree) a tree for a parsed XML input file.
    :param scenarioName: (str) the name of the scenario to apply
    :return: none
    """
    _logger.info("Applying protection scenario %s", scenarioName)

    scenario = Scenario.getScenario(scenarioName)
    if not scenario:
        raise FileFormatError("Scenario '%s' was not found" % scenarioName)

    prot_dict = {}

    for reg, protReg in scenario.protRegDict.iteritems():
        for prot in protReg.protections:
            fraction = prot.fraction
            prot_dict[reg] = [(landtype, fraction) for landtype in prot.landClasses]

    _protect_land(tree, prot_dict)



# def save_unmgd_land_allocation(outfile):
#     """
#     Save a CSV file with reference allocations to all the unmanaged land categories:
#     'Shrubland', 'Grassland', 'UnmanagedPasture', and 'UnmanagedForest'. The original
#     CSV file has GLU (geographical land unit) and region ID keys that we map to their
#     string values for use in XML files.
#     """
#     import pandas as pd
#
#     data_sys_dir = pathjoin(getParam('GCAM.RefWorkspace'), 'input', 'gcam-data-system')
#
#     def _readcsv(path, skiprows=0):
#         filename = data_sys_dir + path
#         df = pd.read_csv(filename, skiprows=skiprows)
#         return df
#
#     df = _readcsv('/aglu-data/level1/L125.LC_bm2_R_LT_Yh_GLU.csv', skiprows=5)
#
#     regionMap = _readcsv('/_common/mappings/GCAM_region_names_32reg.csv', skiprows=3)
#     regionMap.columns = ['ID', 'region']
#     regionMap.set_index('ID', inplace=True)
#     region = regionMap.region
#
#     basinMapFull = _readcsv('/water-data/mappings/basin_to_country_mapping.csv', skiprows=2)
#     basinSlice = basinMapFull[['GLU_code', 'GLU_name']]
#     basinSlice.set_index('GLU_code', inplace=True)
#     basinMap = basinSlice.GLU_name
#
#     years = range(1700,1951, 50) + [1975, 1990, 2005, 2010]
#     xyears = ["X{}".format(year) for year in years]
#     cols = ['GCAM_region_ID', 'Land_Type', 'GLU'] + xyears
#     result = df.query("Land_Type in {}".format(UnmanagedLandClasses))
#     unmgd = result[cols].copy()
#
#     unmgd['region'] = unmgd['GCAM_region_ID'].map(region)
#     unmgd['basin']  = unmgd['GLU'].map(basinMap)
#     unmgd.drop(['GCAM_region_ID', 'GLU'], axis=1, inplace=True)
#
#     unmgd.columns = [col[1:] if col[0] == 'X' else col for col in unmgd.columns]
#     unmgd.rename(index=str, columns={'Land_Type': 'landtype'}, inplace=True)
#
#     unmgd.set_index(['region', 'landtype', 'basin'], inplace=True)
#     unmgd.to_csv(outfile)
#
# def load_unmgd_land_allocation(pathname):
#     """
#     Load the land allocation CSV file saved by `save_unmgd_land_allocation`.
#     """
#     import pandas as pd
#     df = pd.read_csv(pathname, index_col=['region', 'landtype', 'basin'])
#     return df



#!/usr/bin/env python
#
# Installs gcam v4.3 on Windows, Mac OS X, and Linux
#
# Author: Rich Plevin (rich@plevin.com)
# Created: 10 Nov 2016
#
from __future__ import print_function
import os
import sys
import argparse
import subprocess
import platform
import urllib
from urlparse import urlparse
import tarfile

PlatformName = platform.system()
isWindows = (PlatformName == 'Windows')
isMacOS   = (PlatformName == 'Darwin')
isLinux   = (PlatformName == 'Linux')


Home = os.environ['HOME'] = os.environ['HOME']
DefaultInstallDir  = os.path.join(Home, 'gcam-v4.3-install-dir')
DefaultDownloadDir = os.path.join(Home, '.gcam-installation-tmp')

def parseArgs():
    parser = argparse.ArgumentParser(description='''Install GCAM v4.3 on Windows, macOS, or Linux''')

    parser.add_argument('-d', '--downloadDir', default=DefaultDownloadDir,
                        help='''The directory into which to download the required tar files. Default is %s''' % DefaultDownloadDir)

    parser.add_argument('-i', '--installDir', default=DefaultInstallDir,
                        help='''The directory into which to install GCAM 4.3. Default is %s''' % DefaultInstallDir)

    parser.add_argument('-k', '--keepTarFiles', action='store_true',
                        help='''Keep the downloaded tar files rather than deleting them.''')

    parser.add_argument('-n', '--noRun', action='store_true',
                        help='''Print commands that would be executed, but don't run them.''')

    parser.add_argument('-r', '--reuseTarFiles', action='store_true',
                        help='''Use the already-downloaded tar files rather then retrieving them again.
                        Implies -k/--keepTarFiles.''')

    args = parser.parse_args()
    return args


def run(cmd, ignoreError=False, printOnly=False):
    print(cmd)
    if printOnly:
        return

    status = subprocess.call(cmd, shell=True)

    if not ignoreError and status != 0:
        sys.exit(status)

def download(url, filename, printOnly=False):
    def _report(blocks, blockSize, fileSize):
        print('\r%2.0f%% ...' % (100. * blocks * blockSize / fileSize), end='')

    if not filename:
        obj = urlparse(url)
        filename = os.path.basename(obj.path)

    print('Download "%s" -> "%s"\n' % (url, filename))
    if printOnly:
        return

    filename, headers = urllib.urlretrieve(url, filename=filename, reporthook=_report)
    print('') # for the newline
    return filename

def untar(filename, directory='.', printOnly=False):
    print("untar '%s' in '%s'" % (filename, directory))
    if printOnly:
        return

    with tarfile.open(filename, mode='r:gz') as tar:
        tar.extractall(path=directory)

#
# Comment taken from exe/run-gcam.command:
#
# We need to find where the Java development kit is installed.
# This could be the Apple supplied version which was provided up
# to 1.6 however was dropped subsequently and instead users may
# have an Oracle provided JDK.  The each take slightly different
# approaches to where libraries live and how to reference them so
# we will have to try to detect the appropriate location.
#
def fixMacJava(coreDir, printOnly=False):
    cmd = '/usr/libexec/java_home'
    javaHome = ''
    try:
        javaHome = subprocess.check_output(cmd).strip()
    except Exception:
        pass

    if not javaHome:
        print('ERROR: Could not find Java install location')
        sys.exit(-1)

    # If javaHome contains "1.6", use the Apple supplied version of java 1.6
    libPath = 'lib-stub' if '1.6' in javaHome else javaHome + '/jre/lib/server'

    owd = os.getcwd()
    if not printOnly:
        os.chdir(coreDir)

    # Create a symlink to satisfy @rpath searches
    linkName = 'libs/java/lib'
    if not os.path.islink(linkName):
        run("ln -s %s %s" % (libPath, linkName), printOnly=printOnly)

    os.chdir(owd)


def main():
    args = parseArgs()

    coreTarFile = 'gcam-v4.3.tar.gz'
    dataTarFile = 'data-system.tar.gz'

    downloadPrefix = 'https://github.com/JGCRI/gcam-core/releases/download/gcam-v4.3/'

    coreURL = 'https://github.com/JGCRI/gcam-core/archive/' + coreTarFile
    dataURL = downloadPrefix + dataTarFile

    tarFiles = [coreTarFile, dataTarFile]

    madeDownloadDir = False

    downloadDir = os.path.abspath(args.downloadDir)
    installDir  = os.path.abspath(args.installDir)
    printOnly   = args.noRun

    if not os.path.lexists(downloadDir):
        run('mkdir ' + downloadDir, printOnly=printOnly)
        madeDownloadDir = True

    elif not os.path.isdir(downloadDir):
        print('Specified download dir is not a directory: %s' % downloadDir)
        return -1

    startDir = os.getcwd()

    if not printOnly:
        os.chdir(downloadDir)

    if not args.reuseTarFiles:
        download(coreURL, coreTarFile, printOnly=printOnly)
        download(dataURL, dataTarFile, printOnly=printOnly)

    if not os.path.isdir(installDir):
        run('mkdir ' + installDir, printOnly=printOnly)

    coreDir = os.path.join(installDir, 'gcam-core-gcam-v4.3')

    untar(os.path.join(downloadDir, coreTarFile), installDir, printOnly=printOnly)
    untar(os.path.join(downloadDir, dataTarFile), coreDir,    printOnly=printOnly)

    if isMacOS or isWindows:
        binTarFile = '%s_binaries.tar.gz' % ('mac' if isMacOS else 'windows')
        binURL = downloadPrefix + binTarFile

        if not args.reuseTarFiles:
            download(binURL, binTarFile, printOnly=printOnly)

        untar(os.path.join(downloadDir, binTarFile), coreDir, printOnly=printOnly)
        tarFiles.append(binTarFile)

        if isMacOS:
            fixMacJava(coreDir, printOnly=printOnly)

    if printOnly:
        return 0

    print("Installed GCAM into %s" % installDir)

    if args.reuseTarFiles or args.keepTarFiles:
        print("Keeping tar files in %s" % downloadDir)
    else:
        for tarFile in tarFiles:
            os.remove(tarFile)

        if madeDownloadDir:
            # only remove this dir if we created it
            os.chdir(startDir)
            os.rmdir(downloadDir)

    return 0

sys.exit(main())



'''
.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from os import path
from .log import getLogger
from .xmlEditor import XMLEditor, xmlEdit, extractStubTechnology, callableMethod, ENERGY_TRANSFORMATION_TAG

_logger = getLogger(__name__)

# TBD: move to constants.py?
REFINING_SECTOR = 'refining'
BIOMASS_LIQUIDS = 'biomass liquids'

# Technology names
TECH_CORN_ETHANOL       = 'corn ethanol'
TECH_SUGARCANE_ETHANOL  = 'sugar cane ethanol'
TECH_CELLULOSIC_ETHANOL = 'cellulosic ethanol'
TECH_FT_BIOFUELS        = 'FT biofuels'
TECH_BIODIESEL          = 'biodiesel'
TECH_GTL                = 'gas to liquids'
TECH_CTL                = 'coal to liquids'
TECH_OIL_REFINING       = 'oil refining'

ENERGY_SUPPLY_TAG = "energy_supply"
RESBIO_INPUT_TAG  = "residue_bio"
RESOURCES_TAG     = "resources"
AG_BASE_TAG       = "ag_base"
LAND_INPUT3_TAG   = "land3"

class BioenergyEditor(XMLEditor):
    """
    BioenergyEditor adds knowledge of biomass and biofuels.
    """
    def __init__(self, baseline, scenario, xmlOutputRoot, xmlSourceDir, workspaceDir,
                 groupDir, srcGroupDir, subdir, parent=None):
        super(BioenergyEditor, self).__init__(baseline, scenario, xmlOutputRoot, xmlSourceDir,
                                              workspaceDir, groupDir, srcGroupDir, subdir,
                                              parent=parent)

        # TBD: unclear whether this is useful or general
        cornEthanolUsaFile = 'cornEthanolUSA.xml'
        self.cornEthanolUsaAbs = path.join(self.scenario_dir_abs, cornEthanolUsaFile)
        self.cornEthanolUsaRel = path.join(self.scenario_dir_rel, cornEthanolUsaFile)

        cornEthanolUsaFile2 = 'cornEthanolUSA2.xml'
        self.cornEthanolUsaAbs2 = path.join(self.scenario_dir_abs, cornEthanolUsaFile2)
        self.cornEthanolUsaRel2 = path.join(self.scenario_dir_rel, cornEthanolUsaFile2)

        cellEthanolUsaFile = 'cellEthanolUSA.xml'
        self.cellEthanolUsaAbs = path.join(self.scenario_dir_abs, cellEthanolUsaFile)
        self.cellEthanolUsaRel = path.join(self.scenario_dir_rel, cellEthanolUsaFile)

        ftBiofuelsUsaFile = 'ftBiofuelsUSA.xml'
        self.ftBiofuelsUsaAbs = path.join(self.scenario_dir_abs, ftBiofuelsUsaFile)
        self.ftBiofuelsUsaRel = path.join(self.scenario_dir_rel, ftBiofuelsUsaFile)

        # A US subsidy works without having to change prices, so no need to extract this
        biodieselUsaFile = 'biodieselUSA.xml'
        self.biodieselUsaAbs = path.join(self.scenario_dir_abs, biodieselUsaFile)
        self.biodieselUsaRel = path.join(self.scenario_dir_rel, biodieselUsaFile)

        biodieselUsaFile2 = 'biodieselUSA2.xml'
        self.biodieselUsaAbs2 = path.join(self.scenario_dir_abs, biodieselUsaFile2)
        self.biodieselUsaRel2 = path.join(self.scenario_dir_rel, biodieselUsaFile2)

    @callableMethod
    def adjustResidueSupply(self, loTarget, loPrice, loFract, hiTarget, hiPrice, hiFract, target):
        """
        Change supply curves for residues, as per arguments. loTarget and hiTarget identify
        the price to match in the XML file; loPrice and hiPrice are the new prices to set;
        loFract and hiFract are the new fractions to assign to these prices.
        Target must be one of {all, us-crops, us-corn, global-corn}
        Note that the standard reference supply curve for residue is:
        0% at 1975$0; 25% at $1.2; 65% at $1.5; 100% at $10.
        """
        _logger.info("Adjust residue supply curves for %s (%s, %s:%s@%s, %s:%s@%s)" % \
             (self.name, target, loTarget, loFract, loPrice, hiTarget, hiFract, hiPrice))

        # Create modified version of resbio_input.xml and modify config to use it
        resbioFileRel, resbioFileAbs = self.getLocalCopy(RESBIO_INPUT_TAG)

        # Change all non-forest residue, all non-forest residue in the US, only corn residue in US, or corn residue everywhere.
        if target == 'all-crops':
            xPrefix = "//region/AgSupplySector[@name!='Forest']/AgSupplySubsector"

        elif target == 'us-crops':
            xPrefix = "//region[@name='USA']/AgSupplySector[@name!='Forest']/AgSupplySubsector"

        elif target == 'us-corn':
            xPrefix = "//region[@name='USA']/AgSupplySector[@name='Corn']/AgSupplySubsector"

        elif target == 'global-corn':
            xPrefix = "//region/AgSupplySector[@name='Corn']/AgSupplySubsector"

        fractHarvested = xPrefix + "/AgProductionTechnology/period[@year>2010]/residue-biomass-production/fract-harvested"

        pairs = [(fractHarvested + "[@price='%s']" % loTarget, loFract),
                 (fractHarvested + "[@price='%s']" % hiTarget, hiFract),
                 (fractHarvested + "[@price='%s']/@price" % loTarget, loPrice),
                 (fractHarvested + "[@price='%s']/@price" % hiTarget, hiPrice)]

        xmlEdit(resbioFileAbs, pairs)
        self.updateScenarioComponent("residue_bio", resbioFileRel)

    @callableMethod
    def setMswParameter(self, region, parameter, value):
        resourcesFileRel, resourcesFileAbs = self.getLocalCopy(RESOURCES_TAG)

        xpath = "//region[@name='%s']/renewresource/smooth-renewable-subresource[@name='generic waste biomass']/%s" % (region, parameter)

        xmlEdit(resourcesFileAbs, [(xpath, value)])

        self.updateScenarioComponent("resources", resourcesFileRel)

    @callableMethod
    def regionalizeBiomassMarket(self, region):
        _logger.info("Regionalize %s biomass market for %s" % (region, self.name))

        resourcesFileRel, resourcesFileAbs = self.getLocalCopy(RESOURCES_TAG)

        xmlEdit(resourcesFileAbs, [("//region[@name='%s']/renewresource[@name='biomass']/market" % region, region)])
        self.updateScenarioComponent("resources", resourcesFileRel)

        agForPastBioFileRel, agForPastBioFileAbs = self.getLocalCopy(AG_BASE_TAG)

        xmlEdit(agForPastBioFileAbs, [("//region[@name='%s']/AgSupplySector[@name='biomass']/market" % region, region)])
        self.updateScenarioComponent("ag_base", agForPastBioFileRel)

    # TBD: generalize this as setInputCoefficients(self, (('wholesale gas', x), ('elect_td_ind', y)))
    @callableMethod
    def setCornEthanolCoefficients(self, cornCoef, gasCoef=None, elecCoef=None):
        '''
        Set corn ethanol performance coefficients, i.e., the (regional) corn, gas, and
        electricity required per GJ of ethanol. These appear in two files, so we edit
        them both.
        '''
        _logger.info("Set global corn ethanol coefficients: corn=%s, gas=%s, elec=%s" % \
             (cornCoef, gasCoef, elecCoef))

        # XPath applies to file energy-xml/en_supply.xml
        cornCoefXpath = '//technology[@name="regional corn for ethanol"]/period[@year>=2015]/minicam-energy-input[@name="Corn"]/coefficient'

        enSupplyFileRel, enSupplyFileAbs = self.getLocalCopy(ENERGY_SUPPLY_TAG)

        xmlEdit(enSupplyFileAbs, [(cornCoefXpath, cornCoef)])

        self.updateScenarioComponent(ENERGY_SUPPLY_TAG, enSupplyFileRel)

        if gasCoef or elecCoef:
            enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)
            pairs = []
            xpath = '//technology[@name="corn ethanol"]/period[@year>=2015]/minicam-energy-input[@name="%s"]/coefficient'

            if gasCoef:
                gasCoefXpath  = xpath % 'wholesale gas'
                pairs.append((gasCoefXpath, gasCoef))

            if elecCoef:
                elecCoefXpath = xpath % 'elect_td_ind'
                pairs.append((elecCoefXpath, elecCoef))

            xmlEdit(enTransFileAbs, pairs)
            self.updateScenarioComponent("energy_transformation", enTransFileRel)

    # deprecated?
    @callableMethod
    def setBiofuelBiomassCoefficients(self, fuelName, pairs):
        '''
        Set new coefficients for biomass conversion for the given fuel

        :param fuelName:
            The name of the liquid fuel, e.g., 'cellulosic ethanol', 'FT biofuel', etc.
        :param pairs:
            A sequence of tuples of the form (year, coefficient). For example, to set
            the coefficients for cellulosic ethanol for years 2020 and 2025 to 1.234,
            the pairs would be ((2020, 1.234), (2025, 1.234)).
        :return:
            nothing
        '''
        self.setEnergyTechnologyCoefficients(self, BIOMASS_LIQUIDS, fuelName, 'regional biomass', pairs)

    # TBD: generalize as setIsNewTechnology() ?
    @callableMethod
    def purposeGrownOffInRegion(self, region):
        '''
        Turn off the "isNewTechnology" flag for all land-leaf nodes to turn off
        purpose-grown biomass. The line(s) we need to edit in land_input_3.xml is:
        <isNewTechnology fillout="1" year="2020">1</isNewTechnology>
        '''
        _logger.info("Turn off purpose-grown biomass technology in %s for %s" % (region, self.name))

        landInput3Rel, landInput3Abs = self.getLocalCopy(LAND_INPUT3_TAG)

        if region == 'global':
             region='*'

        xmlEdit(landInput3Abs, [("//region[@name='%s']//isNewTechnology[@year='2020']" % region, 0)])
        self.updateScenarioComponent(LAND_INPUT3_TAG, landInput3Rel)

    #
    # Various methods that operate on the USA specifically
    #
    # TBD: make region an optional parameter in these; fix callers
    #
    @callableMethod
    def adjustForestResidueSupply(self, region, loPrice, loFract, hiPrice, hiFract):
        '''
        Change supply curves for forest residues in the USA only. loPrice and hiPrice
        are the new prices to set; loFract and hiFract are the new fractions to assign
        to these prices.
        '''
        _logger.info("Adjust forest residue supply curves for %s" % self.name)

        # Create modified version of resbio_input.xml and modify config to use it
        resbioFileRel, resbioFileAbs = self.getLocalCopy(RESBIO_INPUT_TAG)

        # Forest residue appears in two places. First, operate on AgSupplySector "Forest"
        xPrefix = "//region[@name='%s']/AgSupplySector[@name='Forest']/AgSupplySubsector" % region
        fractHarvested = xPrefix + "/AgProductionTechnology/period[@year>=2015]/residue-biomass-production/fract-harvested"

        pairs = [(fractHarvested + "[@price='1.2']/@price", loPrice),
                 (fractHarvested + "[@price='1.5']/@price", hiPrice),
                 (fractHarvested + "[@price='1.2']", loFract),
                 (fractHarvested + "[@price='1.5']", hiFract)]
        xmlEdit(resbioFileAbs, pairs)

        # Do the same for supplysector="NonFoodDemand_Forest"
        xPrefix = "//region[@name='%s']/supplysector[@name='NonFoodDemand_Forest']/subsector[@name='Forest']/stub-technology[@name='Forest']" % region
        fractHarvested = xPrefix + "/period[@year>=2015]/residue-biomass-production/fract-harvested"

        pairs = [("%s[@price='1.2']/@price" % fractHarvested, loPrice),
                 ("%s[@price='1.5']/@price" % fractHarvested, hiPrice),
                 ("%s[@price='1.2']" % fractHarvested, loFract),
                 ("%s[@price='1.5']" % fractHarvested, hiFract)]
        xmlEdit(resbioFileAbs, pairs)

        self.updateScenarioComponent("residue_bio", resbioFileRel)

    #
    # Methods to operate in USA only on technologies extracted from global-technology-database
    #
    # TBD: generalize this?
    @callableMethod
    def setCellEthanolShareWeightUSA(self, year, shareweight):
        '''
        Create modified version of cellEthanolUSA.xml with the given share-weight
        for the given fuel in the given year.
        '''
        _logger.info("Set US share-weight to %s for cellulosic ethanol in %s for %s" % (shareweight, year, self.name))

        yearConstraint = ">= 2015" if year == 'all' else ("=" + year)

        xmlEdit(self.cellEthanolUsaAbs,
                [("//stub-technology[@name='cellulosic ethanol']/period[@year%s]/share-weight" % yearConstraint,
                  shareweight)])

        self.updateScenarioComponent("cell-etoh-USA", self.cellEthanolUsaRel)

    @callableMethod
    def localizeCornEthanolTechnologyUSA(self):
        '''
        Copy the stub-technology for corn ethanol to CornEthanolUSA.xml and CornEthanolUSA2.xml
        so they can be manipulated in the US only.
        '''
        _logger.info("Add corn ethanol stub technology in USA")

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)
        extractStubTechnology('USA', enTransFileAbs, self.cornEthanolUsaAbs,  REFINING_SECTOR, BIOMASS_LIQUIDS, 'corn ethanol')
        extractStubTechnology('USA', enTransFileAbs, self.cornEthanolUsaAbs2, REFINING_SECTOR, BIOMASS_LIQUIDS, 'corn ethanol', fromRegion=True)

        # Insert "2" right after energy_transformation, then "1" right after energy_transformation,
        # so they end up in order "1" then "2".
        self.insertScenarioComponent('corn-etoh-USA2', self.cornEthanolUsaRel2, 'energy_transformation')
        self.insertScenarioComponent('corn-etoh-USA1', self.cornEthanolUsaRel,  'energy_transformation')

    @callableMethod
    def localizeCellEthanolTechnologyUSA(self):
        '''
        Same as corn ethanol above, but for cellulosic ethanol
        '''
        _logger.info("Add cellulosic ethanol stub-technology in USA")

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)
        extractStubTechnology('USA', enTransFileAbs, self.cellEthanolUsaAbs,  REFINING_SECTOR, BIOMASS_LIQUIDS, 'cellulosic ethanol')

        self.insertScenarioComponent('cell-etoh-USA', self.cellEthanolUsaRel, 'energy_transformation')

    @callableMethod
    def localizeFtBiofuelsTechnologyUSA(self):
        '''
        Same as cellulosic ethanol above
        '''
        _logger.info("Add FT biofuels stub-technology in USA")

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)
        extractStubTechnology('USA', enTransFileAbs, self.ftBiofuelsUsaAbs,  REFINING_SECTOR, BIOMASS_LIQUIDS, 'FT biofuels')

        self.insertScenarioComponent('FT-biofuels-USA', self.ftBiofuelsUsaRel, 'energy_transformation')

    # TBD: translate pairs to allow ranges to be specified
    @callableMethod
    def setBiofuelRefiningNonEnergyCostUSA(self, fuel, pairs):
        if not pairs:
            return

        pathMap = {'corn ethanol'       : self.cornEthanolUsaAbs,
                   'cellulosic ethanol' : self.cellEthanolUsaAbs,
                   'FT biofuels'        : self.ftBiofuelsUsaAbs}

        fuels = pathMap.keys()

        assert fuel in fuels, 'setBiofuelRefiningNonEnergyCostUSA: Fuel must be one of %s' % fuels

        _logger.info("Set US %s non-energy-cost to %s" % (fuel, pairs))

        prefix = "//stub-technology[@name='%s']" % fuel
        suffix = "/minicam-non-energy-input[@name='non-energy']/input-cost"

        pairs = []
        for year, price in pairs:
           pairs.append((prefix + ("/period[@year='%s']" % year) + suffix, price))

        abspath = pathMap[fuel]
        xmlEdit(abspath, pairs)

    @callableMethod
    def setCornEthanolCoefficientsUSA(self, cornCoef, gasCoef=None, elecCoef=None):
        '''
        Set corn ethanol performance coefficients: (regional) corn, gas, and electricity
        required per GJ of ethanol. Modified from superclass version to operate on
        biofuelTechUSA.xml.
        '''
        _logger.info("Set US corn ethanol coefficients for %s (corn=%s, gas=%s, elec=%s)" % \
             (self.name, cornCoef, gasCoef, elecCoef))

        # Corn input is in elements extracted from global-technology-database in energy-xml/en_supply.xml
        cornCoefXpath = '//stub-technology[@name="regional corn for ethanol"]/period[@year>=2015]/minicam-energy-input[@name="Corn"]/coefficient'

        pairs = [(cornCoefXpath, cornCoef)]

        xpath = '//stub-technology[@name="corn ethanol"]/period[@year>=2015]/minicam-energy-input[@name="%s"]/coefficient'
        if gasCoef:
            gasCoefXpath  = xpath % 'wholesale gas'
            pairs.append((gasCoefXpath,  gasCoef))

        if elecCoef:
            elecCoefXpath = xpath % 'elect_td_ind'
            pairs.append((elecCoefXpath,  elecCoef))

        xmlEdit(self.cornEthanolUsaAbs, pairs)
        # config update handled in localize...()

    @callableMethod
    def setCellEthanolBiomassCoefficientsUSA(self, tuples):

        _logger.info("Set US cellulosic ethanol biomass coefficients for %s" % self.name)

        prefix = "//stub-technology[@name='cellulosic ethanol']"
        suffix = "minicam-energy-input[@name='regional biomass']/coefficient"

        pairs = []
        for year, coef in tuples:
            pairs.append(("%s/period[@year='%s']/%s" % (prefix, year, suffix), coef))

        xmlEdit(self.cellEthanolUsaAbs, pairs)
        # config update handled in localize...()



# Copyright (c) 2017 Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from .McsSubcommandABC import McsSubcommandABC

class ExploreCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Run the MCS "explorer" app in dash.'''}
        super(ExploreCommand, self).__init__('explore', subparsers, kwargs,
                                             guiSuppress=True)  # explore runs its own dash server, separate from "gui"

    def addArgs(self, parser):
        parser.add_argument('-d', '--debug', action='store_true',
                            help='''Enable debug mode in the dash server''')
        parser.add_argument('-H', '--host', default='127.0.0.1',
                            help='''Set the host address to serve the application on. Default is localhost (127.0.0.1).''')
        parser.add_argument('-P', '--port', default=8050, type=int,
                            help='''Set the port to serve the application on. Default is 8050.''')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        from ..explorer import main
        main(args)

PluginClass = ExploreCommand



'''
.. Created on: 8/21/16

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from collections import OrderedDict
import glob
import os
import sys

from .config import getParam, getParamAsBoolean
from .error import PygcamException, SetupException
from .log import getLogger
from .utils import getBooleanXML, resourceStream, symlinkOrCopyFile, pathjoin
from .xmlEditor import XMLEditor, getCallableMethod, CachedFile
from .XMLFile import XMLFile, McsValues

_logger = getLogger(__name__)

_tab = ' ' * 3

# Return a class (or any object) defined in this module
def classForString(className):
    return getattr(sys.modules[__name__], className)

def _classForNode(node):
    className = node.tag.capitalize()
    cls = classForString(className)
    return cls(node)

def iterateList(scenarioSetup, cls, node, expandFunc, iterators):
    """
    Recursively evaluate iterators for generalized nested loop, adding
    values to the templateDict in scenarioSetup. Expand ScenarioGroup
    at the inner-most loop, when recursion ends.
    """
    iterName = iterators[0]
    otherIters = iterators[1:]
    iterator = scenarioSetup.getIterator(iterName)
    templateDict = scenarioSetup.templateDict

    for value in iterator.values:
        templateDict[iterName] = value

        if otherIters:
            iterateList(scenarioSetup, cls, node, expandFunc, otherIters)
        else:
            obj = cls(node)
            expandFunc(obj)

#
# Classes to parse and run scenario setup files.
# (See pygcam/etc/scenarios-schema.xsd).
#
class ScenarioSetup(object):

    def __init__(self, node):
        self.name = node.get('name', '')    # unused currently
        self.defaultGroup = node.get('defaultGroup')

        # These serve as a no-op on the build-out pass. The directory vars are
        # converted when the scenario is run and the directories are known.
        self.templateDict = {'scenarioDir' : '{scenarioDir}',
                             'baselineDir' : '{baselineDir}'}

        self.iterators = map(Iterator, node.findall('iterator'))
        self.iteratorDict = {obj.name : obj for obj in self.iterators}

        templateGroups = map(ScenarioGroup, node.findall('scenarioGroup'))

        # Create a dict of expanded groups for lookup
        self.groupDict = OrderedDict()
        self.expandGroups(templateGroups)   # saves into groupDict

    def scenariosInGroup(self, groupName=None):
        group = self.groupDict[groupName or self.defaultGroup]
        return group.scenarioNames()

    def baselineForGroup(self, groupName=None):
        group = self.groupDict[groupName or self.defaultGroup]
        return group.baseline

    def getIterator(self, name):
        try:
            return self.iteratorDict[name]
        except KeyError:
            raise SetupException("Iterator '%s' is not defined" % name)

    documentCache = {}

    @classmethod
    def parse(cls, filename):
        """
        Parse an XML file with query descriptions.
        :param filename: (str) the name of the XML file to read
        :return: a QueryFile instance.
        """
        if filename in cls.documentCache:
            _logger.debug('Found scenario file "%s" in cache', filename)
            return cls.documentCache[filename]

        xmlFile = XMLFile(filename, schemaPath='etc/scenarios-schema.xsd', conditionalXML=True)
        obj = cls(xmlFile.getRoot())

        cls.documentCache[filename] = obj      # cache it
        return obj

    def writeXML(self, stream, indent=0):
        stream.write(_tab * indent + '<setup>\n')

        # No need to show these once expanded
        # for obj in self.iterators:
        #     obj.writeXML(stream, indent + 1)

        for obj in self.groupDict.values():
            obj.writeXML(stream, indent + 1)

        stream.write(_tab * indent + '</setup>\n')

    def expandGroups(self, templateGroups):
        '''
        Expand the `templateGroups`, which may contain names based
        on iterators, into final scenarioGroups without iterators.
        Recursively expands scenarios within groups.
        '''
        templateDict = self.templateDict

        def expand(group):
            group.name = group.name.format(**templateDict)
            group.expandScenarios(self, templateDict)
            self.groupDict[group.name] = group

        for templateGroup in templateGroups:
            iterName = templateGroup.iteratorName

            if not iterName:
                expand(templateGroup)
                continue

            # allow iterator name to be comma-delimited list of iterators
            iterators = map(str.strip, iterName.split(','))
            iterateList(self, ScenarioGroup, templateGroup.node, expand, iterators)

    def run(self, editor, directoryDict, dynamic=False):
        """
        Run the setup for the given XmlEditor subclass.

        :param editor: (XmlEditor) an instance of a subclass of XmlEditor
        :param directoryDict: (dict) directory with values for {scenarioDir}
            {baselineDir}
        :param dynamic: (bool) if True, run only "dynamic" actions; else
           run only static (non-dynamic) actions.
        :return: none
        """
        self.editor = editor
        group = self.groupDict[editor.groupName or self.defaultGroup]
        scenario = group.getFinalScenario(editor.scenario or editor.baseline)

        _logger.debug('Running %s setup for scenario %s', 'dynamic' if dynamic else 'static', scenario.name)
        scenario.run(editor, directoryDict, dynamic=dynamic)

# Iterators for float and int that *included* the stop value.
# That is, terminal condition is "<= stop", not "< stop" as
# with the standard Python range() function.
def frange(start, stop, step=1.0):
    while start <= stop:
        yield start
        start += step

def irange(start, stop, step=1):
    while start <= stop:
        yield start
        start += step

class Iterator(object):
    def __init__(self, node):
        self.name = node.get('name')
        self.min  = self.max = self.step = self.values = self.format = None

        typeName  = node.get('type', 'list')
        isNumeric = typeName in ('int', 'float')
        self.type = iterType = eval(typeName)   # N.B. schema ensures numeric values

        if isNumeric:
            minValue = node.get('min')
            maxValue = node.get('max')
            if not (minValue and maxValue):
                raise SetupException('%s iterator must provide min and max attributes' % typeName)

            self.min  = iterType(minValue)
            self.max  = iterType(maxValue)
            self.step = iterType(node.get('step', 1))
            self.format = node.get('format', '%d' if self.type == int else '%.1f')

            rangeFunc = irange if self.type == int else frange
            self.values = map(lambda value: self.format % value, rangeFunc(self.min, self.max, self.step))

        else: # 'list'
            valuesStr = node.get('values')
            if not valuesStr:
                raise SetupException('list iterator must provide a values attribute')

            self.values = map(str.strip, valuesStr.split(','))

    def __str__(self):
        desc = "<iterator name='%s' type='%s' " % (self.name, self.type.__name__)

        if self.type == list:
            values = ','.join(self.values)
            desc += "values='%s'/>" % values
        else:
            desc += "min='%s' max='%s' step='%s'/>" % (self.min, self.max, self.step)

        return desc

    def writeXML(self, stream, indent=0):
        stream.write(_tab * indent + "%s\n" % self)

class ScenarioGroup(object):
    def __init__(self, node):
        self.node = node
        self.name = node.get('name')
        self.srcGroupDir = node.get('srcGroupDir', '')  # TBD: fix XML docs to refer to this name
        self.useGroupDir = bool(self.srcGroupDir) or getBooleanXML(node.get('useGroupDir', default='0'))
        self.isDefault = getBooleanXML(node.get('default', default='0'))
        self.iteratorName = node.get('iterator')
        self.baselineSource = node.get('baselineSource')
        self.templateScenarios = scenarios = map(Scenario, node.findall('scenario'))
        self.templateDict = {obj.name: obj for obj in scenarios}
        self.finalDict = OrderedDict()
        self.baseline = None

    def getFinalScenario(self, name):
        try:
            return self.finalDict[name]
        except KeyError:
            raise PygcamException('Scenario "%s" was not found in group "%s"' % (name, self.name))

    def scenarioNames(self):
        return self.finalDict.keys()

    def expandScenarios(self, scenarioSetup, templateDict):
        # Replace the text context in all action elements with expanded version

        # This converts only the iterators. The directories {scenarioDir}
        # and {baselineDir} are converted when the scenario is run.
        def expand(scenario):
            scenario.name = name = scenario.name.format(**templateDict)
            subdir = scenario.node.get('subdir', default=scenario.name)
            scenario.subdir = subdir.format(**templateDict)

            self.finalDict[name] = scenario
            scenario.formatContent(templateDict)
            if scenario.isBaseline:
                if self.baseline:
                    raise SetupException('Group %s declares multiple baselines' % self.name)
                self.baseline = name

        for templateScenario in self.templateScenarios:
            iterName = templateScenario.iteratorName

            if not iterName:
                expand(templateScenario)
                continue

            # allow iterator name to be comma-delimited list of iterators
            iterators = map(str.strip, iterName.split(','))
            iterateList(scenarioSetup, Scenario, templateScenario.node, expand, iterators)

    def writeXML(self, stream, indent=0):
        stream.write('\n')
        stream.write(_tab * indent + '<scenarioGroup name="%s" useGroupDir="%s" srcGroupDir="%s">\n' % \
                     (self.name, int(self.useGroupDir), self.srcGroupDir))

        for obj in self.finalDict.values():
            obj.writeXML(stream, indent + 1)

        stream.write(_tab * indent + '</scenarioGroup>\n')

class Scenario(object):
    def __init__(self, node):
        self.node = node
        self.name = node.get('name')
        self.isBaseline = getBooleanXML(node.get('baseline', default=0))
        self.isActive   = getBooleanXML(node.get('active',   default='1'))
        self.iteratorName = node.get('iterator')
        self.actions = map(_classForNode, node)
        self.subdir = node.get('subdir', default=self.name)

    def __str__(self):
        return "<scenario name='%s'>" % self.name

    def run(self, editor, directoryDict, dynamic=False):
        for action in self.actions:
            action.run(editor, directoryDict, dynamic=dynamic)

    def formatContent(self, templateDict):
        # This converts only the iterators. The directories {scenarioDir}
        # and {baselineDir} are converted when the scenario is run.
        for action in self.actions:
            action.formatContent(templateDict)

    def writeXML(self, stream, indent=0):
        stream.write(_tab * indent + '<scenario name="%s" baseline="%s">\n' % \
                     (self.name, int(self.isBaseline)))

        for obj in self.actions:
            obj.writeXML(stream, indent + 1)

        stream.write(_tab * indent + '</scenario>\n')

class ConfigActionBase(object):
    def __init__(self, node):
        self.tag  = node.tag
        self.content = node.text
        self.formattedContent = None
        self.dynamic = False

    def formatContent(self, formatDict):
        content = self.formattedContent or self.content
        self.formattedContent = content.format(**formatDict) if content else None

    def writeXML(self, stream, indent=0):
        stream.write(_tab * indent + "%s\n" % self)

class ConfigAction(ConfigActionBase):
    def __init__(self, node):
        super(ConfigAction, self).__init__(node)
        self.name = node.get('name')
        self.dynamic = getBooleanXML(node.get('dynamic', '0'))

    def __str__(self):
        tag = self.tag
        content = self.formattedContent or self.content
        return "<%s name='%s'>%s</%s>" % (tag, self.name, content, tag)

    def run(self, editor, directoryDict, dynamic=False):
        if self.dynamic == dynamic:
            self.formatContent(directoryDict)
            self._run(editor)

class Insert(ConfigAction):
    def __init__(self, node):
        super(Insert, self).__init__(node)
        self.after = node.get('after')

    def __str__(self):
        tag = self.tag
        content = self.formattedContent or self.content
        after = " after='%s'" % self.after if self.after else ''
        return "<%s name='%s'%s>%s</%s>" % (tag, self.name, after, content, tag)

    def _run(self, editor):
        editor.insertScenarioComponent(self.name, self.formattedContent, self.after)

class Add(ConfigAction):
    def _run(self, editor):
        editor.addScenarioComponent(self.name, self.formattedContent)

class Replace(ConfigAction):
    def _run(self, editor):
        editor.updateScenarioComponent(self.name, self.formattedContent)

class Delete(ConfigAction):
    def _run(self, editor):
        editor.deleteScenarioComponent(self.name)

    def __str__(self):
        return "<%s name='%s'/>" % (self.tag, self.name)

class Function(ConfigAction):
    def _run(self, editor):
        name = self.name
        method = getCallableMethod(name)
        if not method:
            raise SetupException("<function name='%s'>: function doesn't exist or is not callable from XML" % name)

        args = "(%s)" % self.formattedContent if self.formattedContent else "()"
        codeStr = "editor.%s%s" % (name, args)
        try:
            eval(codeStr)
        except SyntaxError as e:
            raise SetupException("Failed to evaluate expression %s: %s" % (codeStr, e))

    def __str__(self):
        tag = self.tag
        content = self.formattedContent or self.content
        return "<%s name='%s' dynamic='%s'>%s</%s>" % (tag, self.name, self.dynamic, content, tag)

class If(ConfigActionBase):
    def __init__(self, node):
        super(If, self).__init__(node)
        self.value1 = node.get('value1')
        self.value2 = node.get('value2')
        self.matches = getBooleanXML(node.get('matches', '1'))
        self.actions = map(_classForNode, node)
        self.formattedValue1 = ''
        self.formattedValue2 = ''

    def __str__(self):
        value1 = self.formattedValue1 or self.value1
        value2 = self.formattedValue2 or self.value2
        return "<%s value1='%s' value2='%s' matches='%s'/>" % (self.tag, value1, value2, self.matches)

    def writeXML(self, stream, indent=0):
        value1 = self.formattedValue1 or self.value1
        value2 = self.formattedValue2 or self.value2

        # output active actions, without the "<if>"
        values = self.formattedValue2.split(',')
        if (self.formattedValue1 in values) == self.matches:
            for obj in self.actions:
                obj.writeXML(stream, indent)

    # N.B. Override superclass method since this runs regardless of dynamic flag
    def run(self, editor, directoryDict, dynamic=False):
        values = map(str.strip, self.formattedValue2.split(','))
        if (self.formattedValue1 in values) == self.matches:
            for action in self.actions:
                action.run(editor, directoryDict, dynamic=dynamic)

    def formatContent(self, formatDict):
        value1 = self.formattedValue1 or self.value1
        value2 = self.formattedValue2 or self.value2

        self.formattedValue1 = value1.format(**formatDict)
        self.formattedValue2 = value2.format(**formatDict)

        for obj in self.actions:
            obj.formatContent(formatDict)

MCSVALUES_FILE = 'mcsValues.xml'

def createXmlEditorSubclass(setupFile, mcsMode=None):
    """
    Generate a subclass of the given `superclass` that runs the
    XML setup file given by variable GCAM.ScenarioSetupFile.
    If defined, GCAM.ScenarioSetupClass must be of the form:
    "/path/to/module/dir;module.ClassName]". If the variable
    GCAM.ScenarioSetupClass is empty, the class XMLEditor is
    subclassed directly.

    :param setupFile: (str) the pathname of an XML setup file
    :param mcsMode (str) must be 'trial', 'gensim', or None
    :return: (class) A subclass of the given `superclass`
    """
    setupClass = getParam('GCAM.ScenarioSetupClass')
    if setupClass:
        try:
            modPath, dotSpec = setupClass.split(';', 1)
        except Exception:
            raise SetupException('GCAM.ScenarioSetupClass should be of the form "/path/to/moduleDirectory:module.ClassName", got "%s"' % setupClass)

        try:
            from .utils import importFromDotSpec
            sys.path.insert(0, modPath)
            superclass = importFromDotSpec(dotSpec)

        except PygcamException as e:
            raise SetupException("Can't load setup class '%s' from '%s': %s" % (dotSpec, modPath, e))

        # check that superclass inherits from or is an XMLEditor
        assert issubclass(superclass, XMLEditor), 'GCAM.ScenarioSetupClass must be a subclass of pygcam.XMLEditor'
    else:
        superclass = XMLEditor

    class XmlEditorSubclass(superclass):
        def __init__(self, baseline, scenario, xmlOutputRoot, xmlSrcDir, refWorkspace, groupName,
                     srcGroupDir, subdir, parent=None):
            self.parentConfigPath = None

            self.scenarioSetup = scenarioSetup = ScenarioSetup.parse(setupFile) #if parent else None
            group = scenarioSetup.groupDict[groupName or scenarioSetup.defaultGroup]

            # if not a baseline, create a baseline instance as our parent
            if scenario:
                # TBD: test this in FCIP case where baseline builds on FuelShock
                baselineSubdir = None
                baseXmlOutputRoot = pathjoin(os.path.dirname(xmlOutputRoot), baseline)
                parent = XMLEditor(baseline, None, baseXmlOutputRoot, xmlSrcDir, refWorkspace, groupName,
                                   srcGroupDir, baselineSubdir)

            super(XmlEditorSubclass, self).__init__(baseline, scenario, xmlOutputRoot, xmlSrcDir,
                                                    refWorkspace, groupName, srcGroupDir, subdir, parent=parent)
            self.paramFile = None

            # Read shocks from mcsValues.xml if present
            #if self.parent and mcsMode:
            if mcsMode == 'trial':
                # ../../trial-xml/local-xml/base-0/mcsValues.xml
                self.trial_xml_abs = pathjoin(self.xmlOutputRoot, '../trial-xml', normpath=True)
                self.trial_xml_rel = pathjoin('../..', 'trial-xml')

                scenName = self.parent.name if self.parent else self.name
                self.paramFile = pathjoin(self.trial_xml_abs, 'local-xml', self.groupDir,
                                          scenName, MCSVALUES_FILE, normpath=True)

            self.directoryDict = {'scenarioDir': self.scenario_dir_rel,
                                  'baselineDir': self.baseline_dir_rel}

        def setupDynamic(self, args):
            self.groupName = args.group

            super(XmlEditorSubclass, self).setupDynamic(args)

            if self.mcsMode == 'trial':             # TBD: was just "if self.mcsMode" -- test this
                paramFile = self.paramFile
                if paramFile and os.path.lexists(paramFile):
                    self.mcsValues = McsValues(paramFile)

            assert self.scenarioSetup, "XmlEditorSubclass.setupDynamic() was called without having read an XML scenario file"
            self.scenarioSetup.run(self, self.directoryDict, dynamic=True)

            # Add symlinks to any files that were added in the dynamic setup
            dynDir  = self.scenario_dyn_dir_abs
            scenDir = self.scenario_dir_abs
            xmlFiles = glob.glob("%s/*.xml" % scenDir)

            if xmlFiles:
                mode = 'Copy' if getParamAsBoolean('GCAM.CopyAllFiles') else 'Link'
                _logger.info("%s additional static XML files in %s to %s", mode, scenDir, dynDir)
                for xml in xmlFiles:
                    base = os.path.basename(xml)
                    dst = pathjoin(dynDir, base)
                    src = pathjoin(scenDir, base)
                    if not os.path.lexists(dst):
                        symlinkOrCopyFile(src, dst)

            CachedFile.decacheAll()

        def setupStatic(self, args):
            self.groupName = args.group
            scenarioSetup = self.scenarioSetup

            # TBD: test this in FCIP case where baseline builds on FuelShock
            if not self.parent:
                # Before calling setupStatic, we set the parent if there is
                # a declared baseline source. This assumes it is in this
                # project, in a different group directory.
                group = scenarioSetup.groupDict[self.groupName or scenarioSetup.defaultGroup]
                baselineSource = group.baselineSource
                if baselineSource:
                    try:
                        groupName, baselineName = baselineSource.split('/')
                    except ValueError:
                        raise SetupException(
                            'baselineSource error: "%s"; should be of the form "groupDir/baselineDir"' % baselineSource)

                    parentGroup = scenarioSetup.groupDict[groupName]
                    scenario = parentGroup.getFinalScenario(baselineName)
                    if scenario.isBaseline:
                        self.parent = XmlEditorSubclass(baselineName, None, self.xmlOutputRoot, self.xmlSourceDir,
                                                        self.refWorkspace, groupName, group.srcGroupDir, scenario.subdir)
            directoryDict = self.directoryDict

            # not an "else" since parent may be set in "if" above
            if self.parent:
                # patch the template dictionary with the dynamically-determined baseline dir
                directoryDict['baselineDir'] = self.baseline_dir_rel = self.parent.scenario_dir_rel

            super(XmlEditorSubclass, self).setupStatic(args)

            # We add this to the baseline. It's ignored by GCAM, but used by MCS. It needs
            # to be found in the config file to be able to apply distributions to the values.
            # TBD: (old) Note that this was "if self.mcsMode and not self.parent:"
            # TBD: (new) No longer add this to config since GCAM 4.3 doesn't ignore it!
            # TBD:       Current approach is to set e.g., name="../local-xml/baseline/mcsValues.xml"
            # TBD:       in parameters.xml, which avoids lookup by tag in config.xml.
            # mcsValuesPath = pathjoin(self.scenario_dir_abs, MCSVALUES_FILE)
            # if self.mcsMode == 'gensim' and not self.parent and os.path.lexists(mcsValuesPath):
            #     self.addScenarioComponent('mcsValues', pathjoin(self.scenario_dir_rel, MCSVALUES_FILE))

            scenarioSetup.run(self, directoryDict, dynamic=False)
            CachedFile.decacheAll()

    return XmlEditorSubclass



# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

# python 3 compatibility version
from six.moves import xrange
import os
from pygcam.config import getParam, setParam
from pygcam.log import getLogger
from pygcam.utils import mkdirs
from ..context import getSimDir
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def genSALibData(trials, method, paramFileObj, args):
    from ..error import PygcamMcsUserError
    from ..sensitivity import DFLT_PROBLEM_FILE, Sobol, FAST, Morris # , MonteCarlo
    from pygcam.utils import ensureExtension, removeTreeSafely, mkdirs, pathjoin

    SupportedDistros = ['Uniform', 'LogUniform', 'Triangle', 'Linked']

    outFile = args.outFile or os.path.join(getSimDir(args.simId), 'data.sa')
    outFile = ensureExtension(outFile, '.sa')

    if os.path.lexists(outFile):
        # Attempt to mv an existing version of the file to the same name with '~'
        backupFile = outFile + '~'
        if os.path.isdir(backupFile):
            removeTreeSafely(backupFile)

        elif os.path.lexists(backupFile):
            raise PygcamMcsUserError("Refusing to delete '%s' since it's not a file package." % backupFile)

        os.rename(outFile, backupFile)

    mkdirs(outFile)

    linked = []

    problemFile = pathjoin(outFile, DFLT_PROBLEM_FILE)
    with open(problemFile, 'w') as f:
        f.write('name,low,high\n')

        for elt in paramFileObj.tree.iterfind('//Parameter'):
            name = elt.get('name')
            distElt = elt.find('Distribution')
            distSpec = distElt[0]   # should have only one child as per schema
            distName = distSpec.tag

            # These 3 distro forms specify min and max values, which we use with SALib
            legal = SupportedDistros + ['Linked']
            if distName not in legal:
                raise PygcamMcsUserError("Found '%s' distribution; must be one of %s for use with SALib." % (distName, legal))

            if distName == 'Linked':        # TBD: ignore these and figure it out when loading the file?
                linked.append((name, distSpec.get('parameter')))
                continue

            # Parse out the various forms: (max, min), (factor), (range), and factor for LogUniform
            attrib = distSpec.attrib
            if 'min' in attrib and 'max' in attrib:
                minValue = float(attrib['min'])
                maxValue = float(attrib['max'])
            elif 'factor' in attrib:
                value = float(attrib['factor'])
                if distName == 'LogUniform':
                    minValue = 1/value
                    maxValue = value
                else:
                    minValue = 1 - value
                    maxValue = 1 + value
            elif 'range' in attrib:
                value = float(attrib['range'])
                minValue = -value
                maxValue =  value

            f.write("%s,%f,%f\n" % (name, minValue, maxValue))

    methods = (Sobol, FAST, Morris) # , MonteCarlo)
    methodMap = {cls.__name__.lower(): cls for cls in methods}

    cls = methodMap[method]
    sa = cls(outFile)

    # saves to input.csv in file package
    sa.sample(trials=trials, calc_second_order=args.calcSecondOrder)
    return sa.inputsDF


def genTrialData(simId, trials, paramFileObj, args):
    """
    Generate the given number of trials for the given simId, using the objects created
    by parsing parameters.xml. Return a DataFrame of values.
    """
    from pandas import DataFrame
    from ..distro import linkedDistro
    from ..LHS import lhs, lhsAmend
    from ..XMLParameterFile import XMLRandomVar, XMLCorrelation
    from ..util import writeTrialDataFile

    rvList = XMLRandomVar.getInstances()

    linked = filter(lambda obj: obj.param.dataSrc.isLinked(), rvList)

    method = args.method
    if method == 'montecarlo':
        # legacy Monte Carlo method. Supporting numerous distributions and correlations.

        corrMatrix = XMLCorrelation.corrMatrix()

        # TBD: this is currently broken. Only "shared" RVs are supported with the DataFrame
        # TBD: approach since a Parameter can have multiple RVs which appear as separate
        # TBD: data columns. This is ok for the current analysis, but needs to be rewritten
        # TBD: on integration with pygcam. (getName() will fail on XMLVariable instances)

        paramNames = map(lambda obj: obj.getParameter().getName(), rvList)
        trialData = lhs(rvList, trials, corrMat=corrMatrix, columns=paramNames, skip=linked)
        writeTrialDataFile(simId, trialData)
    else:
        # SALib methods
        trialData = genSALibData(trials, method, paramFileObj, args)

    linkedDistro.storeTrialData(trialData)  # so its ppf() can access linked values
    lhsAmend(trialData, linked, trials)

    df = DataFrame(data=trialData)
    return df


def saveTrialData(df, simId, start=0):
    """
    Save the trial data in `df` to the SQL database, for the given simId.
    """
    from ..Database import getDatabase
    from ..XMLParameterFile import XMLRandomVar

    trials = df.shape[0]

    # Delete all Trial entries for this simId and this range of trialNums
    db = getDatabase()

    paramValues = []

    for trial in xrange(trials):
        trialNum = trial + start

        # Save all RV values to the database
        instances = XMLRandomVar.getInstances()
        for var in instances:
            varNum = var.getVarNum()
            param = var.getParameter()
            pname = param.getName()
            value = df[pname][trial]
            paramId = db.getParamId(pname)
            paramValues.append((trialNum, paramId, value, varNum))

    # Write the tuples (simId, trialNum, paramId, value) to the database
    db.saveParameterValues(simId, paramValues)

    # SALib methods may not create exactly the number of trials requested
    # so we update the database to set the record straight.
    db.updateSimTrials(simId, trials)
    _logger.info('Generated %d trials for simId %d', trials, simId)


def runStaticSetup(runWorkspace, project, groupName):
    """
    Run the --staticOnly setup in the MCS copy of the workspace, for all scenarios.
    This is called from gensim, so we fake the "context" for trial 0, since all
    trials have local-xml symlinked to RunWorkspace's local-xml.
    """
    import pygcam.tool
    from pygcam.utils import mkdirs
    from pygcam.constants import LOCAL_XML_NAME
    from ..util import symlink
    from ..error import GcamToolError

    projectName = project.projectName

    scenarios = project.getKnownScenarios()
    scenariosArg = ','.join(scenarios)

    useGroupDir = project.scenarioGroup.useGroupDir
    groupSubdir = groupName if useGroupDir else ''

    # create symlinks from all the scenarios' local-xml dirs to shared one
    # under {projectName}/Workspace
    sandboxDir = os.path.join(runWorkspace, groupSubdir)
    mkdirs(sandboxDir)

    wsXmlDir = os.path.join(runWorkspace, LOCAL_XML_NAME)
    mkdirs(wsXmlDir)

    for scenario in scenarios:
        dirname = os.path.join(sandboxDir, scenario)
        mkdirs(dirname)
        linkname  = os.path.join(dirname, LOCAL_XML_NAME)
        symlink(wsXmlDir, linkname)

    # N.B. RunWorkspace for gensim is pygcam's RefWorkspace
    toolArgs = ['+P', projectName, '--mcs=gensim', 'run', '-s', 'setup',
                '-S', scenariosArg, '--sandboxDir=' + sandboxDir]

    if useGroupDir:
        toolArgs += ['-g', groupName]

    _logger.debug('Running: %s', 'gt ' + ' '.join(toolArgs))
    status = pygcam.tool.main(argv=toolArgs, raiseError=True)

    if status != 0:
        msg = '"gt setup" exited with status %d' % status
        raise GcamToolError(msg)

    return status


def genSimulation(simId, trials, paramPath, args):
    '''
    Generate a simulation based on the given parameters.
    '''
    from ..context import Context
    from ..Database import getDatabase
    from ..XMLParameterFile import XMLParameterFile
    from ..util import getSimParameterFile, getSimResultFile, symlink, filecopy
    from pygcam.constants import LOCAL_XML_NAME
    from pygcam.project import Project
    from pygcam.xmlSetup import ScenarioSetup

    runInputDir = getParam('MCS.RunInputDir')
    runWorkspace = getParam('MCS.RunWorkspace')

    # Add symlink to workspace's input dir so we can find XML files using rel paths in config files
    simDir = getSimDir(simId, create=True)
    simInputDir = os.path.join(simDir, 'input')
    symlink(runInputDir, simInputDir)

    # Ditto for workspace's local-xml
    workspaceLocalXml = os.path.join(runWorkspace, LOCAL_XML_NAME)
    simLocalXmlDir = os.path.join(simDir, LOCAL_XML_NAME)
    symlink(workspaceLocalXml, simLocalXmlDir)

    projectName = getParam('GCAM.ProjectName')
    project = Project.readProjectFile(projectName, groupName=args.groupName)

    args.groupName = groupName = args.groupName or project.scenarioSetup.defaultGroup

    # Run static setup for all scenarios in the given group
    runStaticSetup(runWorkspace, project, groupName)

    # TBD: Use pygcam scenario def and copy pygcam files, too
    scenarioFile = getParam('GCAM.ScenarioSetupFile')
    scenarioSetup = ScenarioSetup.parse(scenarioFile)
    scenarioNames = scenarioSetup.scenariosInGroup(groupName)
    baseline      = scenarioSetup.baselineForGroup(groupName)

    # Copy the user's results.xml file to {simDir}/app-xml
    userResultFile = getParam('MCS.ResultsFile')
    simResultFile = getSimResultFile(simId)
    mkdirs(os.path.dirname(simResultFile))
    filecopy(userResultFile, simResultFile)

    paramFileObj = XMLParameterFile(paramPath)
    context = Context(projectName=args.projectName, simId=simId, groupName=groupName)
    paramFileObj.loadInputFiles(context, scenarioNames, writeConfigFiles=True)

    if not trials:
        _logger.warn("Simulation meta-data has been copied.")
        return

    # Define the experiments (scenarios) in the database
    db = getDatabase()
    db.addExperiments(scenarioNames, baseline, scenarioFile)

    paramFileObj.generateRandomVars()

    _logger.info("Generating %d trials to %r", trials, simDir)
    df = genTrialData(simId, trials, paramFileObj, args)

    # Save generated values to the database for post-processing
    saveTrialData(df, simId)

    # Also save the param file as parameters.xml, for reference only
    simParamFile = getSimParameterFile(simId)
    filecopy(paramPath, simParamFile)


def _newsim(runWorkspace, trials):
    '''
    Setup the app and run directories for a given user app.
    '''
    import pkgutil

    from pygcam.scenarioSetup import copyWorkspace
    from ..Database import getDatabase
    from ..error import PygcamMcsUserError
    from ..XMLResultFile import XMLResultFile

    if not runWorkspace:
        raise PygcamMcsUserError("MCS.RunWorkspace was not set in the configuration file")

    srcDir = getParam('GCAM.RefWorkspace')
    dstDir = runWorkspace

    copyWorkspace(dstDir, refWorkspace=srcDir, forceCreate=True, mcsMode=True)

    if trials:
        db = getDatabase()   # ensures database initialization
        XMLResultFile.addOutputs()

        # Load SQL script to create convenient views
        text = pkgutil.get_data('pygcam', 'mcs/etc/views.sql')
        db.executeScript(text=text)


def driver(args, tool):
    '''
    Generate a simulation. Do generic setup, then call genSimulation().
    '''
    from pygcam.utils import removeTreeSafely
    from ..Database import getDatabase
    from ..util import saveDict

    projectName = args.projectName
    runRoot = args.runRoot
    if runRoot:
        # TBD: write this to config file under [project] section
        setParam('MCS.Root', runRoot, section=projectName)
        _logger.info('Please add "MCS.Root = %s" to your .pygcam.cfg file in the [%s] section.',
                     runRoot, projectName)

    runDir = getParam('MCS.RunDir', section=projectName)

    if args.delete:
        removeTreeSafely(runDir, ignore_errors=False)

    runWorkspace = getParam('MCS.RunWorkspace')

    simId  = args.simId
    trials = args.trials
    desc   = args.desc

    if not os.path.exists(runWorkspace):
        _newsim(runWorkspace, trials)

    # Called with trials == 0 when setting up a local run directory on /scratch
    if trials:
        # The simId can be provided on command line, in which case we need
        # to delete existing parameter entries for this app and simId.
        db = getDatabase()
        simId = db.createSim(trials, desc, simId=simId)

    paramFile = args.paramFile or getParam('MCS.ParametersFile')
    genSimulation(simId, trials, paramFile, args=args)

    if trials:
        # Save a copy of the arguments used to create this simulation
        simDir = getSimDir(simId)
        argSaveFile = '%s/gcamGenSimArgs.txt' % simDir
        saveDict(vars(args), argSaveFile)


class GensimCommand(McsSubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : 'Generate input files for simulations.'}
        super(GensimCommand, self).__init__('gensim', subparsers, kwargs)

    def addArgs(self, parser):
        parser.add_argument('--delete', action='store_true',
                            help='''DELETE and recreate the simulation "run" directory.''')

        parser.add_argument('-D', '--desc', type=str, default='',
                            help='A brief (<= 256 char) description the simulation.')

        parser.add_argument('-g', '--groupName', default='',
                            help='''The name of a scenario group to process.''')

        parser.add_argument('-m', '--method', choices=['montecarlo', 'sobol', 'fast', 'morris'],
                            default='montecarlo',
                            help='''Use the specified method to generate trial data. Default is "montecarlo".''')

        parser.add_argument('-o', '--outFile',
                            help='''For methods other than "montecarlo". The path to a "package 
                            directory" into which SALib-related data are stored.
                            If the filename does not end in '.sa', this extension is added. The file
                            'problem.csv' within the package directory will contain the parameter specs in
                            SALib format. The file inputs.csv is also generated in the file package using
                            the chosen method's sampling method. If an outFile is not specified, a package
                            of the name 'data.sa' is created in the simulation run-time directory.''')

        parser.add_argument('-p', '--paramFile', default=None,
                            help='''Specify an XML file containing parameter definitions.
                            Defaults to the value of config parameter MCS.ParametersFile
                            (currently %s)''' % getParam('MCS.ParametersFile'))

        runRoot = getParam('MCS.Root')
        parser.add_argument('-r', '--runRoot', default=None,
                            help='''Root of the run-time directory for running user programs. Defaults to
                            value of config parameter MCS.Root (currently %s)''' % runRoot)

        parser.add_argument('-S', '--calcSecondOrder', action='store_true',
                            help='''For Sobol method only -- calculate second-order sensitivities.''')

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help='The id of the simulation. Default is 1.')

        parser.add_argument('-t', '--trials', type=int, required=True,
                            help='The number of trials to create for this simulation (REQUIRED)')

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import argparse
import subprocess

from pygcam.config import getParam

def parseArgs():
    defaultSimId = 1
    defaultTrialCount = 10

    parser = argparse.ArgumentParser(description='''Run typical MCS analysis''')

    parser.add_argument('-b', '--baseline',
                        help='''The name of the baseline scenario, which will be run
                        if specified, unless --noRun is also specified''')

    dfltProject = getParam('GCAM.DefaultProject')
    parser.add_argument('-P', '--project', default=dfltProject,
                        help='''The name of the project to run (current default project is "%s"''' % dfltProject)

    parser.add_argument('-t', '--trialCount', type=int, default=defaultTrialCount,
                        help='''The number of trials to create and run''')

    parser.add_argument('-g', '--groupDir',
                        help='''The name of a scenario group to use as a directory name
                        between the project name and the scenario name.''' )

    parser.add_argument('-G', '--dontGensim', action='store_true',
                        help='''Skip the 'gensim' step, but print the command''')

    parser.add_argument('-n', '--noRun', action='store_true',
                        help='''Print commands that would be executed, but don't run them.''')

    parser.add_argument('-q', '--dontQueue', action='store_true',
                        help='''Skip the 'queue' step, but print the command''')

    parser.add_argument('-s', '--simId', type=int, default=defaultSimId,
                        help='The id of the simulation to run. Default is %s.' % defaultSimId)

    args = parser.parse_args()
    return args


def run(cmd, ignoreError=False, printOnly=False):
    print(cmd)
    if printOnly:
        return

    status = subprocess.call(cmd, shell=True)

    if not ignoreError and status != 0:
        sys.exit(status)


def main():
    args = parseArgs()
    projectName = args.project

    if args.noRun:
        args.dontGensim = args.dontQueue = True

    simId = args.simId

    run("gt -P %s newsim" % projectName, printOnly=args.noRun)

    groupFlag = '-g ' + args.groupDir if args.groupDir else ''
    run("gt -P %s gensim -s %d -t %d %s" % (projectName, simId, args.trialCount, groupFlag),
        printOnly=args.dontGensim)

    if args.baseline:
        run("gt -P %s runsim -s %d -S %s" % (projectName, simId, args.baseline), printOnly=args.dontQueue)


main()



"""
.. Support for querying GCAM's XML database and processing results.

.. codeauthor:: Rich Plevin <rich@plevin.com>

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.

"""
from ..subcommand import SubcommandABC

class QueryCommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'fromfile_prefix_chars' : '@',      # use "@" before value to substitute contents of file as arguments
                  'help' : '''Run one or more GCAM database queries by generating and running the
                  named XML queries.'''}

        super(QueryCommand, self).__init__('query', subparsers, kwargs, group='project')


    def addArgs(self, parser):
        parser.add_argument('queryName', nargs='*',
                            help='''A file or files, each holding an XML query to run. (The ".xml" suffix will be
                            added if needed.) If an argument is preceded by the "@" sign, it is read and its
                            contents substituted as the values for this argument. That means you can store queries
                            to run in a file (one per line) and just reference the file by preceding the filename
                            argument with "@".''')

        parser.add_argument('-b', '--batchFile',
                             help='''An XML batch file to run. The file will typically contain 
                             multiple queries. By default, output is written to 
                             {outputDir}/{batchFile basename}.csv. Use '-B' to change this.''')

        parser.add_argument('-B', '--batchOutput', default='',
                             help='''Where to write the output of the XML batch file given by 
                             the '-b' flag. Non-absolute paths are treated as relative to the given outputDir.''')

        parser.add_argument('-d', '--xmldb',
                             help='''The XML database to query (default is computed as
                             {GCAM.SandboxDir}/output/{GCAM.DbFile}. Overrides the -w flag.''')

        parser.add_argument('-D', '--noDelete', action="store_true",
                            help='''Don't delete any temporary file created by extracting a query from a query file. Used
                                    mainly for debugging.''')

        parser.add_argument('-g', '--groupDir', default='',
                            help='''The scenario group directory name, if any. Used with to compute default
                            for --workspace argument.''')

        parser.add_argument('-n', '--noRun', action="store_true",
                            help="Show the command to be run, but don't run it")

        parser.add_argument('-o', '--outputDir',
                             help='Where to output the result (default taken from config parameter "GCAM.OutputDir")')

        parser.add_argument('-p', '--prequery', action="store_true",
                            help='''Generate the XMLDBDriver.properties file and associated batch file to be
                                 run by GCAM when GCAM.BatchMultipleQueries or GCAM.InMemoryDatabase are True.''')

        parser.add_argument('-q', '--queryXmlFile',
                            help='''An XML file holding a list of queries to run, with optional mappings specified to
                            rewrite output. This file has the same structure as the <queries> element in project.xml.''')

        parser.add_argument('-Q', '--queryPath',
                            help='''A semicolon-delimited list of directories or filenames to look in to find query files.
                                    Defaults to value of config parameter GCAM.QueryPath''')

        parser.add_argument('-r', '--regions',
                            help='''A comma-separated list of regions on which to run queries found in query files structured
                                    like Main_Queries.xml. If not specified, defaults to querying all 32 regions.''')

        parser.add_argument('-R', '--regionMap',
                            help='''A file containing tab-separated pairs of names, the first being a GCAM region
                                    and the second being the name to map this region to. Lines starting with "#" are
                                    treated as comments. Lines without a tab character are also ignored. This arg
                                    overrides the value of config variable GCAM.RegionMapFile.''')

        parser.add_argument('-s', '--scenario', default='Reference',
                            help='''The scenario to run the query/queries for (default is "Reference")
                                    Note that this must refers to a scenarios in the XML database.''')

        parser.add_argument('-S', '--rewriteSetsFile',
                            help='''An XML file defining query maps by name (default taken from
                            config parameter "GCAM.RewriteSetsFile")''')

        parser.add_argument('-w', '--workspace', default='',
                            help='''The workspace directory in which to find the XML database.
                                    Defaults computed as {GCAM.SandboxDir}/{groupDir}/{scenario}.
                                    Overridden by the -d flag.''')

        return parser

    def run(self, args, tool):
        from ..query import queryMain

        queryMain(args)



from .chart_plugin import ChartCommand
from .config_plugin import ConfigCommand
from .diff_plugin import DiffCommand
from .gcam_plugin import GcamCommand
from .gui_plugin import GUICommand
from .init_plugin import InitCommand
from .mcs_plugin import MCSCommand
from .mi_plugin import ModelInterfaceCommand
from .new_plugin import NewProjectCommand
from .protect_plugin import ProtectLandCommand
from .query_plugin import QueryCommand
from .run_plugin import RunCommand
from .sandbox_plugin import SandboxCommand
from .setup_plugin import SetupCommand
from .compare_plugin import CompareCommand

BuiltinSubcommands = [ChartCommand, CompareCommand, ConfigCommand, DiffCommand,
                      GcamCommand, GUICommand, InitCommand, MCSCommand,
                      ModelInterfaceCommand, NewProjectCommand, ProtectLandCommand,
                      QueryCommand, RunCommand, SandboxCommand, SetupCommand]



# Created on 10/25/14
#
# @author : Rich Plevin
#
# @Copyright 2015 Richard Plevin

class PygcamMcsException(Exception):
    'Base class for application-related errors.'
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class PygcamMcsUserError(PygcamMcsException):
    'The user provided an incorrect parameter or failed to provide a required one.'
    pass

class PygcamMcsSystemError(PygcamMcsException):
    'Some application-related runtime error.'
    pass

class IpyparallelError(PygcamMcsSystemError):
    pass

class FileExistsError(PygcamMcsSystemError):
    def __init__(self, filename):
        self.message = "File %r already exists" % filename

    # def __str__(self):
    #     return repr(self.message)

class FileMissingError(PygcamMcsSystemError):
    def __init__(self, filename):
        self.message = "File %r is missing" % filename

class ShellCommandError(PygcamMcsSystemError):
    def __init__(self, msg, exitStatus=0):
        self.message = msg
        self.exitStatus = exitStatus

    def __str__(self):
        statusMsg = " exit status %d" % self.exitStatus if self.exitStatus else ""
        return "ShellCommandError: %s%s" % (statusMsg, self.exitStatus)

class GcamToolError(PygcamMcsSystemError):  # TBD: define in pygcam?
    pass

class BaseSpecError(PygcamMcsSystemError):
    filename = ''
    lineNum = 0

    def __init__(self, message):
        if self.filename and self.lineNum:
            self.message = 'File %s, line %i: ' % (self.filename, self.lineNum) + message
        else:
            self.message = message

class DistributionSpecError(BaseSpecError):
    pass



import dash
import flask
import requests

from pygcam.config import usingMCS
from pygcam.log import getLogger
from pygcam.subcommand import SubcommandABC
from pygcam.tool import GcamTool
from pygcam.gui.widgets import Page, PageSet, RootPage, GlobalArgsPage
from pygcam.gui.terminal import Terminal

_logger = getLogger(__name__)

def _showInGUI(cmd):
    subcmd = SubcommandABC.getInstance(cmd)
    return not subcmd.guiSuppress

def commandGroupPage(app, group, default, label=None):
    cmds = GcamTool.pluginGroup(group, namesOnly=True)
    cmds = filter(_showInGUI, cmds)   # eliminate suppressed subcmds

    pages = [Page(app, cmd) for cmd in cmds]
    pageSet = PageSet(group, pages, default)
    page = Page(app, group, label=label, pageSet=pageSet)
    return page

def driver(args):
    from pkg_resources import resource_string

    app = dash.Dash(csrf_protect=False)
    app.config['suppress_callback_exceptions'] = True

    googleFonts = "https://fonts.googleapis.com/css?family=Dosis:regular,semi-bold|Lato"

    serve_locally = True

    if serve_locally:
        # app.css.config.serve_locally = True
        # app.scripts.config.serve_locally = True

        files = ('stylesheet.css', 'googlefonts.css', 'terminal_scroller.js')
        map_to_remote = {'googlefonts.css': googleFonts}

        @app.server.route('/static/<path:path>')
        def serve_static(path):
            _logger.debug('Serving "%s"' % path)

            # Serve only known files
            if not path in files:
                raise Exception('Unknown static file: "{}"'.format(path))

            url = map_to_remote.get(path)
            if url:
                # Download content from corresponding remote URL
                content = requests.get(url).content
            else:
                # Grab files from the gui/static folder in the pygcam package
                resource_path = 'gui/static/' + path
                content = resource_string('pygcam', resource_path)

            mimetype = "text/css" if path.endswith('.css') else "text/javascript"
            response = flask.Response(content, mimetype=mimetype)

            return response

        for filename in files:
            arg_dict = {'external_url' : '/static/' + filename}

            if filename.endswith('.css'):
                app.css.append_css(arg_dict)
            else:
                app.scripts.append_script(arg_dict)
    else:
        # this version uses external files...
        app.css.append_css({"external_url": googleFonts})
        app.css.append_css({"external_url": "https://codepen.io/plevin/pen/MvpeNV.css"})
        app.scripts.append_script({'external_url': 'https://codepen.io/plevin/pen/MvpeNV.js'})

    debug = args.debug
    term = Terminal(updateSeconds=1.5)

    pages = [Page(app, 'run'),
             commandGroupPage(app, 'project', 'new'),
             commandGroupPage(app, 'mcs', 'runsim', label='MCS') if usingMCS() else None,
             commandGroupPage(app, 'utils', 'mi', label='Utilities'),
             GlobalArgsPage(app)]

    RootPage(app, term, filter(None, pages))

    term.registerCallbacks(app)
    app.run_server(debug=debug)

if __name__ == '__main__':
    driver()



import os
import sys
import unittest

from pygcam.windows import IsWindows

def printLink(path):
    islink = os.path.islink(os.path.normpath(path))
    sys.stderr.write("%s islink: %s\n" % (path, islink))
    if islink:
        path = os.readlink(path)
        sys.stderr.write("Link: %s" % path)
        # printLink(path)

class TestSymlinks(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_funcs(self):
        if IsWindows:
            from pygcam.windows import islinkWindows, readlinkWindows
            self.assertEqual(os.path.islink, islinkWindows)
            self.assertEqual(os.readlink, readlinkWindows)

    def test_islink(self):
        if IsWindows:
            p1 = 'C:/Users/rjp/GCAM/current'
            self.assertTrue(os.path.islink(p1), '%s should be seen as a link' % p1)

            p2 = p1 + '/Main_User_Workspace'

            self.assertFalse(os.path.islink(p2), '%s should not be seen as a link' % p2)



'''
.. pygcam's Exception classes

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
class PygcamException(Exception):
    """
    Base class for pygcam Exceptions.
    """
    pass

class FileMissingError(PygcamException):
    """
    Indicate that a required file was not found or not readable.
    """

    def __init__(self, filename, reason):
        self.filename = filename
        self.reason   = reason

    def __str__(self):
        return "Can't read %s: %s" % (self.filename, self.reason)

class FileFormatError(PygcamException):
    """
    Indicate a syntax error in a user-managed file.
    """
    pass

class XmlFormatError(FileFormatError):
    pass

class FileExistsError(PygcamException):
    """
    Raised when trying to write a file that already exists (if not allowed)
    """
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "Refusing to overwrite file: %s" % self.filename

class ConfigFileError(FileFormatError):
    """
    Raised when an error is found in the configuration file ``~/.pygcam.cfg``.
    """
    pass

class CommandlineError(Exception):
    """
    Command-line arguments were missing or incorrectly specified.
    """
    pass

class ProgramExecutionError(PygcamException):
    """
    Raised when attempt to execute a program (e.g., the GCAM model or ModelInterface) fails.
    """
    def __init__(self, command, exitCode=None):
        self.command = command
        self.exitCode = exitCode

    def __str__(self):
        return "Command '%s' failed with exit code %s" % (self.command, self.exitCode)

class GcamError(PygcamException):
    """
    The gcamWrapper detected and error and terminated the model run.
    """
    pass

class GcamSolverError(GcamError):
    """
    The gcamWrapper detected and error and terminated the model run.
    """
    pass

class SetupException(Exception):
    """
    Raised from the setup sub-system.
    """
    pass



#
# Signal handling
#
import signal
from .windows import IsWindows

# Windows has only SIGABRT, SIGFPE, SIGILL, SIGINT, SIGSEGV, or SIGTERM
_sigmap = {
    signal.SIGABRT: 'SIGABRT',
    signal.SIGINT:  'SIGINT',
    signal.SIGTERM: 'SIGTERM',
} if IsWindows else {
    signal.SIGINT:  'SIGINT',
    signal.SIGALRM: 'SIGALRM',
    signal.SIGTERM: 'SIGTERM',
    signal.SIGQUIT: 'SIGQUIT',
}

def _signame(signum):
    return _sigmap[signum] if signum in _sigmap else 'signal %d' % signum


class SignalException(Exception):
    def __init__(self, signum):
        self.signum = signum
        self.signame = _signame(signum)

    def __str__(self):
        return "Process received %s" % self.signame

class AlarmSignalException(SignalException):
    pass

class TimeoutSignalException(SignalException):
    pass

class UserInterruptException(SignalException):
    pass

def raiseSignalException(signum, _frame):
    if not IsWindows and signum == signal.SIGALRM:
        raise AlarmSignalException(signum)

    elif signum == signal.SIGTERM:      # TBD: this is sent by SLURM. Same for PBS??
        raise TimeoutSignalException(signum)

    elif signum == signal.SIGINT:       # control-c by user
        raise UserInterruptException(signum)

    else:
        raise SignalException(signum)


def catchSignals(handler=raiseSignalException):
    for sig in _sigmap:
        signal.signal(sig, handler)



'''
.. The "gt" (gcamtool) commandline program

.. Copyright (c) 2016-2017 Richard Plevin and UC Regents
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import argparse
from glob import glob
import os
import pipes
import re
import subprocess
import sys

from .built_ins import BuiltinSubcommands
from .config import (getParam, getConfig, getParamAsBoolean, getParamAsFloat,
                     setParam, getSection, setSection, getSections,
                     DEFAULT_SECTION, usingMCS, savePathMap)
from .error import PygcamException, ProgramExecutionError, ConfigFileError, CommandlineError
from .log import getLogger, setLogLevels, configureLogs
from .signals import SignalException, catchSignals
from .version import VERSION
from .windows import IsWindows

PROGRAM = 'gt'

# For now, these are not offered as command-line options. Needs more testing.
# BioConstraintsCommand, DeltaConstraintsCommand,

class GcamTool(object):

    # plugin instances by command name
    _plugins = {}

    # cached plugin paths by command name
    _pluginPaths = {}

    @classmethod
    def getPlugin(cls, name):
        if name not in cls._plugins:
            cls._loadCachedPlugin(name)

        return cls._plugins.get(name, None)

    @classmethod
    def _loadCachedPlugin(cls, name):
        # see if it's already loaded
        path = cls._pluginPaths[name]
        tool = cls.getInstance()
        tool.loadPlugin(path)

    @classmethod
    def _cachePlugins(cls):
        '''
        Find all plugins via GCAM.PluginPath and create a dict
        of plugin pathnames keyed by command name so the plugin
        can be loaded on-demand.
        :return: none
        '''
        from .utils import pathjoin

        pluginDirs = cls._getPluginDirs()

        suffix = '_plugin.py'
        suffixLen = len(suffix)

        for d in pluginDirs:
            pattern = pathjoin(d, '*' + suffix)
            for path in glob(pattern):
                basename = os.path.basename(path)
                command = basename[:-suffixLen]
                cls._pluginPaths[command] = path

    _instance = None

    @classmethod
    def getInstance(cls, loadPlugins=True, reload=False):
        """
        Get the singleton instance of the GcamTool class.

        :param loadPlugins: (bool) If true, plugins are loaded (only
           when first allocated).
        :param reload: (bool) If true, a new GcamTool instance is
           created.
        :return: (GcamTool instance) the new or cached instance.
        """
        if reload:
            GcamTool._instance = None
            GcamTool._plugins = {}
            GcamTool._pluginPaths = {}

        if not GcamTool._instance:
            GcamTool._instance = cls(loadPlugins=loadPlugins)

        return GcamTool._instance

    @classmethod
    def pluginGroup(cls, groupName, namesOnly=False):
        objs = filter(lambda obj: obj.getGroup() == groupName, cls._plugins.values())
        result = sorted(map(lambda obj: obj.name, objs)) if namesOnly else objs
        return result

    def __init__(self, loadPlugins=True, loadBuiltins=True):
        from .project import decacheVariables

        # address re-entry issue
        decacheVariables()

        self.mcsMode = ''
        self.shellArgs = None

        self.parser = self.subparsers = None
        self.addParsers()

        # load all built-in sub-commands
        if loadBuiltins:
            map(self.instantiatePlugin, BuiltinSubcommands)

        # If using MCS, load that set of built-ins, too
        if usingMCS():
            from .mcs.built_ins import MCSBuiltins
            map(self.instantiatePlugin, MCSBuiltins)

        # Load external plug-ins found in plug-in path
        if loadPlugins:
            self._cachePlugins()

    def addParsers(self):
        self.parser = parser = argparse.ArgumentParser(prog=PROGRAM, prefix_chars='-+')

        parser.add_argument('+b', '--batch', action='store_true',
                            help='''Run the commands by submitting a batch job using the command
                            given by config variable GCAM.BatchCommand. (Linux only)''')

        parser.add_argument('+B', '--showBatch', action="store_true",
                            help="Show the batch command to be run, but don't run it. (Linux only)")

        parser.add_argument('+D', '--dirmap',
                            help="""A comma-delimited sequence of colon-delimited directory names 
                            of the form "/some/host/path:/a/container/path, /host:cont, ...", 
                            mapping host dirs to their mount point in a docker container.""")

        parser.add_argument('+e', '--enviroVars',
                            help='''Comma-delimited list of environment variable assignments to pass
                            to queued batch job, e.g., -E "FOO=1,BAR=2". (Linux only)''')

        parser.add_argument('+j', '--jobName', default='gt',
                            help='''Specify a name for the queued batch job. Default is "gt".
                            (Linux only)''')

        logLevel = str(getParam('GCAM.LogLevel'))   # so not unicode
        parser.add_argument('+l', '--logLevel',
                            default=logLevel or 'notset',
                            help='''Sets the log level for modules of the program. A default
                                log level can be set for the entire program, or individual 
                                modules can have levels set using the syntax 
                                "module:level, module:level,...", where the level names must be
                                one of {debug,info,warning,error,fatal} (case insensitive).''')

        parser.add_argument('+L', '--logFile',
                            help='''Sets the name of a log file for batch runs. Default is "gt-%%j.out"
                            where "%%j" (in SLURM) is the jobid. If the argument is not an absolute 
                            pathname, it is treated as relative to the value of GCAM.LogDir.''')

        parser.add_argument('+m', '--minutes', type=float, default=getParamAsFloat('GCAM.Minutes'),
                            help='''Set the number of minutes to allocate for the queued batch job.
                            Overrides config parameter GCAM.Minutes. (Linux only)''')

        parser.add_argument('+M', '--mcs', dest='mcsMode', choices=['trial','gensim'],
                            help='''Used only when running gcamtool from pygcam-mcs.''')

        parser.add_argument('+P', '--projectName', metavar='name', default=getParam('GCAM.DefaultProject'),
                            choices=sorted(getSections()),
                            help='''The project name (the config file section to read from),
                            which defaults to the value of config variable GCAM.DefaultProject''')

        parser.add_argument('+q', '--queueName', default=getParam('GCAM.DefaultQueue'),
                            help='''Specify the name of the queue to which to submit the batch job.
                            Default is given by config variable GCAM.DefaultQueue. (Linux only)''')

        parser.add_argument('+r', '--resources', default='',
                            help='''Specify resources for the queued batch command. Can be a comma-delimited
                            list of assignments of the form NAME=value, e.g., -r 'pvmem=6GB'. (Linux only)''')

        parser.add_argument('+s', '--set', dest='configVars', metavar='name=value', action='append', default=[],
                            help='''Assign a value to override a configuration file parameter. For example,
                            to set batch commands to start after a prior job of the same name completes,
                            use --set "GCAM.OtherBatchArgs=-d singleton". Enclose the argument in quotes if
                            it contains spaces or other characters that would confuse the shell.
                            Use multiple --set flags and arguments to set multiple variables.''')

        parser.add_argument('+v', '--verbose', action='store_true',
                            help='''Show diagnostic output''')

        parser.add_argument('--version', action='version', version='%(prog)s-' + VERSION)   # goes to stderr, handled by argparse

        parser.add_argument('--VERSION', action='store_true')   # goes to stdout, but handled by gt

        self.subparsers = self.parser.add_subparsers(dest='subcommand', title='Subcommands',
                               description='''For help on subcommands, use the "-h" flag after the subcommand name''')

    def setMcsMode(self, mode):
        self.mcsMode = mode

    def getMcsMode(self):
        return self.mcsMode

    def instantiatePlugin(self, pluginClass):
        plugin = pluginClass(self.subparsers)
        self._plugins[plugin.name] = plugin

    @staticmethod
    def _getPluginDirs():
        pluginPath = getParam('GCAM.PluginPath')
        if not pluginPath:
            return []

        sep = os.path.pathsep           # ';' on Windows, ':' on Unix
        items = pluginPath.split(sep)

        return items

    def loadPlugin(self, path):
        """
        Load the plugin at `path`.

        :param path: (str) the pathname of a plugin file.
        :return: an instance of the ``SubcommandABC`` subclass defined in `path`
        """
        from .utils import loadModuleFromPath

        def getModObj(mod, name):
            return getattr(mod, name) if name in mod.__dict__ else None

        mod = loadModuleFromPath(path)

        pluginClass = getModObj(mod, 'PluginClass') or getModObj(mod, 'Plugin')
        if not pluginClass:
            raise PygcamException('Neither PluginClass nor class Plugin are defined in %s' % path)

        self.instantiatePlugin(pluginClass)

    def _loadRequiredPlugins(self, argv):
        # Create a dummy subparser to allow us to identify the requested
        # sub-command so we can load the module if necessary.
        parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
        parser.add_argument('-h', '--help', action='store_true')
        parser.add_argument('+P', '--projectName', metavar='name')

        ns, otherArgs = parser.parse_known_args(args=argv)

        # For top-level help, or if no args, load all plugins
        # so the generated help messages includes all subcommands
        if ns.help or not otherArgs:
            map(self._loadCachedPlugin, self._pluginPaths.keys())
        else:
            # Otherwise, load any referenced sub-command
            for command in self._pluginPaths.keys():
                if command in otherArgs:
                    self.getPlugin(command)

    def validateGcamVersion(self):
        from .utils import pathjoin, pushd, parse_version_info
        from .gcam import setJavaPath

        exeDir  = pathjoin(getParam('GCAM.RefWorkspace'), 'exe')
        exeName = getParam('GCAM.Executable')
        exePath = pathjoin(exeDir, exeName)

        if not os.path.lexists(exePath):
            raise ConfigFileError('GCAM executable "%s" was not found.' % exePath)

        # Starting with v4.3, gcam reports its version number
        versionCfg = parse_version_info()
        if versionCfg >= (4, 3):
            versionFile = pathjoin(exeDir, '.version')

            # Check for cached version info
            if os.path.lexists(versionFile):
                with open(versionFile, 'r') as f:
                    versionNum = f.readline().strip()
            else:
                setJavaPath(exeDir)
                with pushd(exeDir):
                    try:
                        cmd = [exePath, '--versionID']
                        versionStr = subprocess.check_output(cmd, shell=False).strip()

                    except subprocess.CalledProcessError:
                        raise ConfigFileError(
                            "Attempt to run '%s --versionID' failed. If you're running GCAM < v4.3, set GCAM.VersionNumber accordingly" % exePath)

                prefix = 'gcam-v'
                if not versionStr.startswith(prefix):
                    raise ConfigFileError('GCAM --versionID "%s" is not the correct format. Should start with "gcam-v"',
                                          versionStr)

                versionNum = versionStr[len(prefix):]

                # cache version number so we don't have to run 'gcam.exe --versionID' every time
                with open(versionFile, 'w') as f:
                    f.write(versionNum + '\n')

            versionExe = parse_version_info(versionNum)
            if (versionCfg.major, versionCfg.minor) != (versionExe.major, versionExe.minor):
                log = getLogger(__name__)
                log.warning("Setting GCAM.VersionNumber = %s to match GCAM version. (Set it in the config file to suppress this message.)", versionNum)
                setParam('GCAM.VersionNumber', versionNum)

    def run(self, args=None, argList=None):
        """
        Parse the script's arguments and invoke the run() method of the
        designated sub-command.

        :param args: an argparse.Namespace of parsed arguments
        :param argList: (list of str) argument list to parse (when called recursively)
        :return: none
        """
        assert args or argList, "GcamTool.run requires either args or argList"

        checkWindowsSymlinks()

        if argList is not None:         # might be called with empty list of subcmd args
            # called recursively
            self._loadRequiredPlugins(argList)
            args = self.parser.parse_args(args=argList)

        else:  # top-level call
            if args.batch:
                args.batch = False

            args.projectName = section = args.projectName or getParam('GCAM.DefaultProject')
            if section:
                 setSection(section)

            if args.subcommand != 'init':
                # After GCAM.DefaultProject is set, we can check gcam executable
                try:
                    self.validateGcamVersion()
                except ConfigFileError as e:
                    _logger = getLogger(__name__)
                    _logger.warning("%s", e)

            logLevel = args.logLevel or getParam('GCAM.LogLevel')
            if logLevel:
                setLogLevels(logLevel)

            configureLogs(force=True)

        # Get the sub-command and run it with the given args
        obj = self.getPlugin(args.subcommand)

        obj.run(args, self)

    # Extracted from runBatch() to be callable from the run sub-command
    # to distribute scenario runs.
    @staticmethod
    def runBatch2(shellArgs, jobName='gt', queueName=None, logFile=None, minutes=None,
                  dependsOn=None, run=True):
        from .utils import mkdirs, pathjoin

        _logger = getLogger(__name__)

        queueName = queueName or getParam('GCAM.DefaultQueue')
        logFile   = logFile   or getParam('GCAM.BatchLogFile', raw=False)
        minutes   = minutes   or getParamAsFloat('GCAM.Minutes')

        batchSystem = getParam('GCAM.BatchSystem')
        known = ('SLURM', 'LSF', 'PBS')
        if batchSystem not in known:
            raise ConfigFileError('GCAM.Scheduler value (%s) is not recognized. Must be one of %s.' % (batchSystem, known))

        # The LSF scheduler needs HH:MM; SLURM needs HH:MM:SS
        format = "%02d:%02d" if batchSystem == 'LSF' else "%02d:%02d:00"
        walltime  = format % (minutes / 60, minutes % 60)

        if logFile:
            logDir = getParam('GCAM.BatchLogDir')
            logFile = os.path.normpath(pathjoin(logDir, logFile))
            mkdirs(os.path.dirname(logFile))

        # TBD: make this an expression eval'd with s.format(jobID=dependsOn)
        # TBD: to support other syntaxes
        format = "-w 'done(%s)'" if batchSystem == 'LSF' else "-d afterok:%s"
        dependencies = format % dependsOn if dependsOn else ''

        scriptCommand = "gt " + ' '.join(shellArgs)

        # This dictionary is applied to the string value of GCAM.BatchCommand, via
        # the str.format method, which must specify options using any of the keys.
        batchArgs = {'scriptFile': scriptCommand,
                     'dependencies' : dependencies,
                     'logFile'   : logFile,
                     'minutes'   : minutes,
                     'walltime'  : walltime,
                     'queueName' : queueName,
                     'partition' : queueName,   # synonym for queue name
                     'jobName'   : jobName}

        batchCmd = getParam('GCAM.BatchCommand')

        try:
            command = batchCmd.format(**batchArgs)

        except KeyError as e:
            raise ConfigFileError('Badly formatted batch command (%s) in config file: %s', batchCmd, e)

        if not run:
            print(command)
            return

        _logger.info('Running: %s', command)
        try:
            jobStr = subprocess.check_output(command, shell=True)
            result = re.search('\d+', jobStr)
            jobId = int(result.group(0)) if result else -1
            return jobId

        except subprocess.CalledProcessError as e:
            raise ProgramExecutionError(command, e.returncode)

        except Exception as e:
            raise PygcamException("Error running command '%s': %s" % (command, e))

    def runBatch(self, shellArgs, run=True):
        import platform

        system = platform.system()
        if system in ['Windows']: # , 'Darwin']:
            system = 'Mac OS X' if system == 'Darwin' else system
            raise CommandlineError('Batch commands are not supported on %s' % system)

        shellArgs = map(pipes.quote, shellArgs)
        args = self.parser.parse_args(args=shellArgs)

        return self.runBatch2(shellArgs, jobName=args.jobName, queueName=args.queueName,
                              logFile=args.logFile, minutes=args.minutes, run=run)

def _getMainParser():
    '''
    Used only to generate documentation by sphinx' argparse, in which case
    we don't generate documentation for project-specific plugins.
    '''
    getConfig(allowMissing=True)
    tool = GcamTool.getInstance(loadPlugins=False)
    return tool.parser


def checkWindowsSymlinks():
    '''
    If running on Windows and GCAM.CopyAllFiles is not set, and
    we fail to create a test symlink, set GCAM.CopyAllFiles to True.
    '''
    from .temp_file import getTempFile

    if IsWindows and not getParamAsBoolean('GCAM.CopyAllFiles'):
        src = getTempFile()
        dst = getTempFile()

        try:
            os.symlink(src, dst)
        except:
            _logger = getLogger(__name__)
            _logger.info('No symlink permission; setting GCAM.CopyAllFiles = True')
            setParam('GCAM.CopyAllFiles', 'True')

def _setDefaultProject(argv):
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
    parser.add_argument('+P', '--projectName', metavar='name')

    ns, _otherArgs = parser.parse_known_args(args=argv)

    section = ns.projectName
    if section:
        setParam('GCAM.DefaultProject', section, section=DEFAULT_SECTION)
        setSection(section)

def _saveDirMap():
    dirMapFile = os.getenv('DIRMAP_PATH')
    if not dirMapFile:
        return

    with open(dirMapFile) as f:
        dirMap = f.read()

    savePathMap(dirMap)

# This parser handles only --VERSION flag.
def _showVersion(argv):
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')
    parser.add_argument('--VERSION', action='store_true')

    ns, otherArgs = parser.parse_known_args(args=argv)

    if ns.VERSION:
        print(PROGRAM + '-' + VERSION)
        sys.exit(0)

def _main(argv=None):
    from .config import userConfigPath

    configPath = userConfigPath()
    if not os.path.lexists(configPath) or os.stat(configPath).st_size == 0:
        argSet = set(argv or sys.argv)
        options = {'init', '-h', '--help', '--version'}
        if argSet.intersection(options):
            # create empty config file just so we can run the "init" sub-command, or help/version options
            open(configPath, 'w').close()
        else:
            raise CommandlineError('\n***\n*** Missing or empty pygcam configuration file %s. Run "gt init" to create it.\n***\n' % configPath)

    getConfig()

    _showVersion(argv)
    _saveDirMap()

    configureLogs()

    _setDefaultProject(argv)

    tool = GcamTool.getInstance()
    tool._loadRequiredPlugins(argv)

    # This parser handles only --batch, --showBatch, --projectName, --set, and --mcs
    # args. If --batch is given, we need to create a script and call the
    # GCAM.BatchCommand on it. We grab --projectName so we can set PluginPath by project.
    parser = argparse.ArgumentParser(prog=PROGRAM, add_help=False, prefix_chars='-+')

    parser.add_argument('+b', '--batch', action='store_true')
    parser.add_argument('+B', '--showBatch', action="store_true")
    parser.add_argument('+P', '--projectName', dest='projectName', metavar='name')
    parser.add_argument('+s', '--set', dest='configVars', action='append', default=[])
    parser.add_argument('+M', '--mcs', dest='mcsMode', choices=['trial','gensim'])

    ns, otherArgs = parser.parse_known_args(args=argv)

    tool.setMcsMode(ns.mcsMode)

    # Set specified config vars
    for arg in ns.configVars:
        if not '=' in arg:
            raise CommandlineError('-S requires an argument of the form variable=value, got "%s"' % arg)

        name, value = arg.split('=')
        setParam(name, value)

    # showBatch => don't run batch command, but implies --batch
    if ns.showBatch:
        ns.batch = True

    # Catch signals to allow cleanup of TempFile instances, e.g., on ^C
    catchSignals()

    if ns.batch:
        run = not ns.showBatch
        if ns.projectName:        # add these back in for the batch script
            otherArgs = ['+P', ns.projectName] + otherArgs

        tool.runBatch(otherArgs, run=run)
    else:
        tool.shellArgs = otherArgs  # save for project run method to use in "distribute" mode
        args = tool.parser.parse_args(args=otherArgs)
        tool.run(args=args)


def main(argv=None, raiseError=False):
    try:
        _main(argv)
        return 0

    except CommandlineError as e:
        print(e)

    except SignalException as e:
        if raiseError:
            raise

        _logger = getLogger(__name__)
        _logger.error("%s: %s" % (PROGRAM, e))
        return e.signum

    except Exception as e:
        if raiseError:
            raise

        print("%s failed: %s" % (PROGRAM, e))

        if not getSection() or getParamAsBoolean('GCAM.ShowStackTrace'):
            import traceback
            traceback.print_exc()

    finally:
        from .temp_file import TempFile

        # Delete any temporary files that were created
        TempFile.deleteAll()

    return 1



from pygcam.mcs.Database import getDatabase
from pygcam.mcs.util import createTrialString

db = getDatabase(checkInit=False)

simId = 1
df = db.getRunInfo(simId, 'baseline', includeSucceededRuns=False, asDataFrame=True)
print df

missing = db.getMissingTrials(simId, 'baseline')
s = createTrialString(missing)
print "Missing trials:", s

from pygcam.mcs.master import listTrialsToRedo

listTrialsToRedo(db, simId, ['baseline', 'corn'], ['missing', 'killed'])



from .addexp_plugin import AddExpCommand
from .analyze_plugin import AnalyzeCommand
from .cluster_plugin import ClusterCommand
from .delsim_plugin import DelSimCommand
from .discrete_plugin import DiscreteCommand
from .engine_plugin import EngineCommand
from .explore_plugin import ExploreCommand
from .gensim_plugin import GensimCommand
from .ippsetup_plugin import IppSetupCommand
from .iterate_plugin import IterateCommand
from .parallelPlot_plugin import ParallelPlotCommand
from .runsim_plugin import RunSimCommand

MCSBuiltins = [AddExpCommand, AnalyzeCommand, ClusterCommand,
               DiscreteCommand, GensimCommand, DelSimCommand,
               EngineCommand, ExploreCommand, IppSetupCommand,
               IterateCommand, ParallelPlotCommand, RunSimCommand]



import platform
from pygcam.version import VERSION

# Build the full version with MCS on when running on ReadTheDocs.
# In normal mode, MCS is an optional install.
import os

if platform.system() != 'Windows':
    # Unfortunately, this stalled on Windows when I tested it...
    import ez_setup
    ez_setup.use_setuptools(version='36.7.2') # ... is this version avail on RTD?

from setuptools import setup

requirements = [
    'configparser>=3.5.0',     # backport of python 3.5 version
    'filelock>=2.0.12',
    'futures>=3.1.1',
    'future>=0.16.0',
    'ipython<6.0',              # 6.x requires Python 3
    'lxml>=3.8.0',
    'numpy>=1.13.1',
    'pandas>=0.20.3',
    'seaborn>=0.8.0',
    'semver>=2.7.7',
    'six>=1.10.0',
    'sphinx-argparse==0.2.2', # later versions lose markup in commands
    'tornado<5.0',
    
    # GUI requirements
    'flask>=0.12.2',
    'dash>=0.19.0',
    'dash-core-components>=0.12.7',
    'dash-html-components>=0.8.0',
    'dash-renderer>=0.7.4',

    # MCS requirements
    'ipyparallel>=6.0.2',
    'numexpr>=2.6.2',
    'salib==1.1.2',
    'scipy>=0.19.1',
    'sqlalchemy>=1.1.13',
]

#if os.environ.get('READTHEDOCS') == 'True':
#    requirements.extend(mcs_requirements)
#    extras_requirements = {}
#else:
#    extras_requirements = {
#        'mcs': mcs_requirements,
#    }

long_description = '''
pygcam
=======

The ``pygcam`` package provides a workflow management framework for GCAM
consisting of scripts and a Python API for customizing the framework by
writing plug-ins, or by writing new applications.

Full documentation and a tutorial are available at
https://pygcam.readthedocs.io.

Core functionality
------------------

* Project workflow management framework that lets you define steps to run and
  run them all or run steps selectively.

* The main ``gt`` ("GCAM tool") script, which provides numerous
  sub-commands, and can be extended by writing plug-ins.

* The ``gt`` sub-commands facilitate key steps in working with GCAM, including:

  * Setting up experiments by modifying XML input files and configuration.xml
  * Running GCAM, locally or by queueing jobs on a Linux cluster
  * Querying the GCAM database to extract results to CSV files
  * Interpolating between time-steps and computing differences between baseline and policy cases
  * Plotting results
  * Setting up, running, and analyzing Monte Carlo simulations on Linux clusters

* The scripts are based on the pygcam API, documented at https://pygcam.readthedocs.io

* Scripts that provide flexible command-line interfaces to the functionality provided by
  the library.

* Customization through an extensive configuration system

How do I get set up?
----------------------

* Users on OS X and Windows platforms can download a zip file with an all-in-one
  directory that has everything you need to run the "gt" (gcamtool) command.

* Linux users and anyone wishing to use ``pygcam`` for Python development should
  install it as a normal Python package. The easiest way to to install directly from
  PyPi:

    ``pip install pygcam``

  Alternatively, clone the repository or download the tarball and run this command
  on the included setup.py file:

    ``python setup.py install``

  or, if you want to edit the code or stay abreast of code changes, you might install
  it in "developer" mode:

    ``python setup.py develop``

Contribution guidelines
------------------------

* TBD

Who do I talk to?
------------------

* Rich Plevin (rich@plevin.com)
'''

setup(
    name='pygcam',
    version=VERSION,
    description='Python 2.7 library and scripts for interfacing with GCAM',
    platforms=['Windows', 'MacOS', 'Linux'],

    packages=['pygcam'],
    entry_points={'console_scripts': ['gt = pygcam.tool:main']},
    install_requires=requirements,
    include_package_data = True,

    # extras_require=extras_requirements,

    url='https://github.com/JGCRI/pygcam',
    download_url='https://github.com/JGCRI/pygcam.git',
    license='MIT License',
    author='Richard Plevin',
    author_email='rich@plevin.com',

    classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: MIT License',
          'Intended Audience :: Science/Research',
          ],

    zip_safe=True,

    # TBD
    # test_suite="string identifying test module etc."

)




# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from .McsSubcommandABC import McsSubcommandABC

def driver(args, tool):
    """
    Start or stop an ipyparallel cluster. If starting the cluster, the number of
    engines launched is determined from the number of tasks to run (`ntasks`) and
    the maximum number of engines indicated in config file param IPP.MaxEngines.
    Wall time allowed per engine is computed from these values and config file
    variable IPP.MinutesPerRun.
    """
    from ..master import startCluster, stopCluster

    if args.mode == 'start':
        argDict = vars(args)
        startCluster(**argDict)
    else:
        stopCluster(profile=args.profile, cluster_id=args.clusterId,
                    stop_jobs=args.stopJobs, other_args=args.otherArgs)


class ClusterCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Start an ipyparallel cluster after generating batch
        file templates based on parameters in .pygcam.cfg and the number of tasks to run.'''}
        super(ClusterCommand, self).__init__('cluster', subparsers, kwargs)

    def addArgs(self, parser):
        from pygcam.config import getParam, getParamAsFloat, getParamAsInt

        defaultProfile     = getParam('IPP.Profile')
        defaultClusterId   = getParam('IPP.ClusterId')
        defaultQueue       = getParam('IPP.Queue')
        defaultWorkDir     = getParam('IPP.WorkDir')
        defaultStopJobsCmd = getParam('IPP.StopJobsCommand')

        defaultMaxEngines = getParamAsInt('IPP.MaxEngines')
        defaultMinutes    = getParamAsFloat('IPP.MinutesPerRun')

        parser.add_argument('mode', choices=['start', 'stop'],
                            help='''Whether to start or stop the cluster''')

        parser.add_argument('-c', '--clusterId', type=str, default=defaultClusterId,
                            help='''A string to identify this cluster. Default is the
                            value of config var IPP.ClusterId, currently
                            "%s".''' % defaultClusterId)

        parser.add_argument('-e', '--maxEngines', type=int, default=defaultMaxEngines,
                            help='''Set maximum number of engines to create.
                            Overrides config parameter IPP.MaxEngines, currently
                            %s''' % defaultMaxEngines)

        parser.add_argument('-m', '--minutesPerRun', type=int, default=defaultMinutes,
                            help='''Set the number of minutes of walltime to allocate
                            per GCAM run.  Overrides config parameter IPP.MinutesPerRun,
                            currently %s.''' % defaultMinutes)

        parser.add_argument('-n', '--numTrials', type=int, default=10,
                            help='''The total number of GCAM trials that will be run on this
                            cluster. (Relevant only for "start" command.)''')

        parser.add_argument('-o', '--otherArgs', type=str, default='',
                            help='Command line arguments to append to the ipcluster command.')

        parser.add_argument('-p', '--profile', type=str, default=defaultProfile,
                            help='''The name of the ipython profile to use. Default is
                            the value of config var IPP.Profile, currently
                            "%s".''' % defaultProfile)

        parser.add_argument('-q', '--queue', type=str, default=defaultQueue,
                            help='''The queue or partition on which to create the controller
                            and engines. Overrides config var IPP.Queue, currently
                            "%s".''' % defaultQueue)

        parser.add_argument('-s', '--stopJobs', action='store_true',
                            help='''Stop running jobs using the value if IPP.StopJobsCommand,
                            currently "%s". (Ignored for mode "start".)''' % defaultStopJobsCmd)

        parser.add_argument('-w', '--workDir', type=str, default=defaultWorkDir,
                            help='''Where to run the ipcluster command. Overrides the
                            value of config var IPP.WorkDir, currently '%s'.''' % defaultWorkDir)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        driver(args, tool)



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 24 18:51:21 2014

@author: Ryan Jones

Copyright (c) 2014 Ryan Jones and Richard Plevin.
See the file COPYRIGHT.txt for details.
"""
import numpy as np
from pygcam.matplotlibFix import plt
import pandas as pd
import seaborn as sns

import tsplotModified as tsm
from .analysis import printExtraText

# TBD: Generalize this when integrating into pygcam
def plotForcingSubplots(tsdata, filename=None, ci=95, show_figure=False, save_fig_kwargs=None):
    sns.set_context('paper')
    expList = tsdata['expName'].unique()

    nrows = 1
    ncols = len(expList)
    width  = 2 * ncols
    height = 2
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, sharey=True, figsize=(width, height))

    def dataForExp(expName):
        df = tsdata.query("expName == '%s'" % expName).copy()
        df.drop(['expName'], axis=1, inplace=True)
        df = pd.melt(df, id_vars=['runId'], var_name='year')
        return df

    for ax, expName in zip(axes, expList):
        df = dataForExp(expName)

        pos = expName.find('-')
        title = expName[:pos] if pos >= 0 else expName
        ax.set_title(title.capitalize())

        tsm.tsplot(df, time='year', unit='runId', value='value', ci=ci, ax=ax)

        ylabel = 'W m$^{-2}$' if ax == axes[0] else ''
        ax.set_ylabel(ylabel)
        ax.set_xlabel('') # no need to say "year"
        ax.axhline(0, color='navy', linewidth=0.5, linestyle='-')
        plt.setp(ax.get_xticklabels(), rotation=270)

    plt.tight_layout()

    # Save the file
    if filename:
        if isinstance(save_fig_kwargs, dict):
            fig.savefig(filename, **save_fig_kwargs)
        else:
            fig.savefig(filename)

    # Display the figure
    if show_figure:
        plt.show()

    return fig


def plotTimeSeries(datasets, timeCol, unit, valueCol='value', estimator=np.mean, estimator_linewidth=1.5,
                   ci=90, legend_loc='upper left', legend_labels=None, legend_name=None, title=None,
                   xlabel=None, ylabel=None, label_font=None, title_font=None, xlim=None, ylim=None,
                   ymin=None, ymax=None, text_label=None, show_figure=True, filename=None,
                   save_fig_kwargs=None, figure_style='darkgrid', palette_name=None, extra=None):
    """
    Plot one or more timeseries with flexible representation of uncertainty.

    This function can take a single ndarray array or a list of ndarray arrays
    and plot them against time (years) with a specified confidence interval.

    Parameters
    ----------
    datasets : ndarray, dataframe, or list of ndarrays or dataframes
        Data for the plot. Rows represent samples, columns represent years
    years : series-like
        x values for a plot when data is an array.
    estimator : function
        Function operates column wise on the datasets to produce each line in the figure
    ci : float or list of floats in [0, 100]
        Confidence interval size(s). If a list, it will stack the error
        plots for each confidence interval. Only relevant for error styles
        with "ci" in the name.
    legend_loc : String or float
        Location of the legend on the figure
    legend_labels : string or list of strings
        Either the name of the field corresponding to the data values in
        the data DataFrame (i.e. the y coordinate) or a string that forms
        the y axis label when data is an array.
    legend_name : string
        Legend title.
    title : string
        Plot title
    xlabel : string
        x axis label
    ylabel : string
        y axis label
    text_label : string or list of strings
        if a list of strings, each string gets put on a separate line
    show_figure : bool
        Boolean indicating whether figure should be shown
    filename : string
        Filename used in saving the figure
    save_fig_kwargs : dict
        Other keyword arguments are passed to savefig() call
    figure_style :
        Seaborn figure background styles, options include:
        darkgrid, whitegrid, dark, white
    palette_name : seaborn palette
        Palette for the main plots and error representation

    Returns
    -------
    fig : matplotlib figure

    """

    # Set plot style
    sns.set_style(figure_style)

    # Set up dataset
    if legend_labels is None:
        legend_labels = [None]*len(datasets)
        legend = False
    else:
        legend = True

    if isinstance(legend_labels, str):
        legend_labels = [legend_labels]

    if isinstance(datasets, np.ndarray) or isinstance(datasets, pd.DataFrame):
        datasets = [datasets]

    # Colors
    #colors = sns.color_palette(name=palette_name, n_colors=len(datasets))      # strangely claims name is not a known keyword.
    colors = sns.color_palette(n_colors=len(datasets))

    # Create the plots
    #if fig is None:
    fig, ax = plt.subplots()

    # TBD: this is probably ok, but shouldn't save if a subplot, i.e., if fig & ax were passed in
    for color, data, series_name in zip(colors, datasets, legend_labels):
        tsm.tsplot(data, time=timeCol, value=valueCol, unit=unit, ci=ci, ax=ax, color=color,
                   condition=series_name, estimator=estimator, linewidth=estimator_linewidth)
        # standard version computes CI with different semantics
        #sns.tsplot(data, time=timeCol, value=valueCol, unit=unit, ci=ci, ax=ax, color=color,
        #           condition=series_name, estimator=estimator, linewidth=estimator_linewidth)

    # Add the plot labels
    if label_font is None:
        label_font = dict()
    label_font.setdefault('size', 'medium')

    if title_font is None:
        title_font = dict()
    title_font.setdefault('size', 'large')

    if xlabel is not None:
        ax.set_xlabel(xlabel, fontdict=label_font)

    if ylabel is not None:
        ax.set_ylabel(ylabel, fontdict=label_font)

    if title is not None:
        ax.set_title(title, fontdict=title_font)

    printExtraText(fig, extra, color='grey', loc='right')

    if text_label is not None:
        axis = ax.axis()
        if not isinstance(text_label, str):
            text_label = '\n'.join(text_label)
        ax.text(axis[0]+(axis[1]-axis[0])*.03, axis[2]+(axis[3]-axis[2])*.7, text_label, fontdict=label_font)

    if legend:
        legend1 = ax.legend(loc=legend_loc, title=legend_name, prop={'size': 'medium'})

    for label in (ax.get_xticklabels() + ax.get_yticklabels()):
        label.set(fontsize='medium')

    # Axis limits
    if xlim is not None:
        ax.set_xlim(xlim[0], xlim[1])

    if ymin is not None or ymax is not None:
        ax.set_autoscale_on(False)
        ax.set_ylim(ymin, ymax)

    elif ylim is not None:
        ax.set_ylim(ylim[0], ylim[1])

    # Save the file
    if filename:
        if isinstance(save_fig_kwargs, dict):
            fig.savefig(filename, **save_fig_kwargs)
        else:
            fig.savefig(filename)

    # Display the figure
    if show_figure:
        plt.show()

    return fig


# if __name__ == "__main__":
#     def test_confidence_interval_plot():
#
#         # Create a fake dataset
#         years = range(1990, 2055, 5)
#         datasets = []
#         datasets.append(np.random.normal(100, 25, (100,len(years)))*(np.arange(1,len(years)+1)**1.1))
#
#         plotTimeSeries(datasets, years, ci=[95, 50], title='Test Figure', xlabel='Year', ylabel='Emissions (Mg CO$_2$)',
#                        legend_labels=['Total CO$_2$ Emissions'], legend_name='Legend', text_label=['95% Confidence Interval','50% Confidence Interval'])
#
#     test_confidence_interval_plot()





'''
.. codeauthor:: Richard Plevin

.. Copyright (c) 2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
from ..subcommand import SubcommandABC

DefaultProperties = '''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE properties SYSTEM "http://java.sun.com/dtd/properties.dtd">
<properties>
    <comment>TODO: add comments</comment>
    <entry key="queryFile">{queryFile}</entry>
</properties>
'''

class ModelInterfaceCommand(SubcommandABC):

    def __init__(self, subparsers):
        kwargs = {'help' : '''Run ModelInterface for the current project.'''}

        super(ModelInterfaceCommand, self).__init__('mi', subparsers, kwargs, group='utils', label='MI')

    def addArgs(self, parser):
        parser.add_argument('-d', '--useDefault', action='store_true',
                            help='''Use the Main_Queries.xml file from the GCAM
                            reference workspace.''')
        parser.add_argument('-u', '--updateProperties', action='store_true',
                            help='''Update the "model_interface.properties" file in the directory
                            indicated by config var file GCAM.QueryDir so it refers to the query file
                            indicated by config var GCAM.MI.QueryFile, or if this does not refer
                            to an existing file, by var GCAM.MI.RefQueryFile.''')
        return parser

    def run(self, args, tool):
        import os
        from ..utils import shellCommand, pushd
        from ..xmlEditor import xmlEdit
        from ..config import getParam
        from ..log import getLogger

        _logger = getLogger(__name__)

        queryFile = getParam('GCAM.MI.QueryFile')
        if args.useDefault or not os.path.lexists(queryFile):
            queryFile = getParam('GCAM.MI.RefQueryFile')

        queryDir = getParam('GCAM.QueryDir')

        propFile = 'model_interface.properties'

        with pushd(queryDir):
            # Create the default file if none exists
            if not os.path.lexists(propFile):
                with open(propFile, 'w') as f:
                    text = DefaultProperties.format(queryFile=queryFile)
                    f.write(text)

            # Update the file if so requested
            if args.updateProperties:
                pairs = [('//entry[@key="queryFile"]', queryFile)]
                xmlEdit(propFile, pairs, useCache=False)

            # run ModelInterface
            cmd = getParam('GCAM.MI.Command')
            _logger.debug(cmd)
            shellCommand(cmd, shell=True, raiseError=True)


PluginClass = ModelInterfaceCommand



from __future__ import print_function
from pygcam.matplotlibFix import plt

import pandas as pd
from pandas.tools.plotting import parallel_coordinates
import seaborn as sns

from .analysis import spearmanCorrelation

# get_ipython().magic(u'matplotlib inline')

Feedstocks = ['corn', 'soy', 'sugarcane', 'grass']

def isLinked(colname):
    pos = colname.find('[')
    colname = colname if pos < 0 else colname[0:pos]
    return colname.endswith('-linked')

def getCorrDF(inputs, output):
    corrDF = pd.DataFrame(spearmanCorrelation(inputs, output))
    corrDF['abs'] = corrDF.spearman.abs()
    corrDF.sort_values('abs', ascending=False, inplace=True)
    return corrDF

def normalize(df):
    result = df.copy()
    for name in df.columns:
        mn = df[name].max()
        mx = df[name].min()
        result[name] = (df[name] - mn) / (mx - mn)
    return result

def topContributors(inputs, output, count):
    corrDF = getCorrDF(inputs, output)
    top = list(corrDF.index[:count])
    return top

def importantParams(df, outputs, count):
    result = {key: topContributors(df, outputs[key], count) for key in Feedstocks}
    return result

def extractCols(df, colnames):
    extracted = df[colnames]
    df.drop(extracted.columns, axis=1, inplace=True)
    return extracted

def flipCorrSign(df, colNames, corrDF):
    df = df.loc[:, colNames]
    for name in colNames:
        if corrDF.spearman[name] < 0:
            df[name] = 1 - df[name]
            df.rename(columns={name: "(1 - %s)" % name}, inplace=True)
    return df

def categorizeCI(count, inputDF, title=None, addCount=False):
    largest = inputDF.nlargest(count, 'ci')
    smallest = inputDF.nsmallest(count, 'ci')
    largest['tag'] = 'High CI'
    smallest['tag'] = 'Low CI'
    joined = pd.concat((largest,smallest))
    joined.drop(['ci'], axis=1, inplace=True)

    g = parallel_coordinates(joined, 'tag', color=[[0.8,0,0.1,0.3],[0,0.1,0.8,0.3]])
    plt.xticks(rotation=270)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    if title:
        if addCount:
            title += ' (n=%d)' % count
        g.set_title(title)
    return g

def categorizeCI2(inputDF, subsampleFactor=10, title=None):
    #inputDF = normalize(inputDF)
    binLabels = ['Low', 'Medium', 'High']
    indices = range(0, inputDF.shape[0], subsampleFactor)
    plotDF = inputDF.iloc[indices].copy()
    plotDF['bin'] = pd.qcut(inputDF['ci'], len(binLabels), labels=binLabels)
    plotDF.drop(['ci'], axis=1, inplace=True)
    alpha = 0.3
    g = parallel_coordinates(plotDF, 'bin',
                             color=[[0.8,0.0,0.1,alpha],
                                    [0.0,0.8,0.1,alpha],
                                    [0.1,0.1,0.8,alpha],
                                   ])
    plt.xticks(rotation=270)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    if title:
        title += ' (factor=%d)' % subsampleFactor
        g.set_title(title)
    return g

def createInputDF(df, topContribs, feedstock):
    inputDF = df[topContribs[feedstock]].copy()
    inputDF['ci'] = ci[feedstock]
    return normalize(inputDF.dropna())


if __name__ == "__main__":
    filename = 'Combined.csv'
    df = pd.read_table(filename, sep=',', index_col='trial')

    linked = filter(isLinked, df.columns)
    if linked:
        df.drop(linked, axis=1, inplace=True)

    ci = extractCols(df, Feedstocks)
    contribs = importantParams(df, ci, 5)

    g = sns.boxplot(ci)
    g.axes.set_title('All data')

    categorizeCI2(createInputDF(df, contribs, 'soy'),
                  subsampleFactor=20,
                  title='Soy biodiesel')

    categorizeCI2(createInputDF(df, contribs, 'corn'),
                  subsampleFactor=10,
                  title='Corn ethanol')

    categorizeCI2(createInputDF(df, contribs, 'grass'),
                  subsampleFactor=20,
                  title='Energy grass ethanol')

    categorizeCI2(createInputDF(df, contribs, 'sugarcane'),
                  subsampleFactor=25,
                  title='Sugarcane ethanol')

    categorizeCI(100, createInputDF(df, contribs, 'corn'), title='Corn ethanol', addCount=True)

    cornCorrDF = getCorrDF(df, ci['corn'])
    inputDF = createInputDF(df, contribs, 'corn')
    inputDF = flipCorrSign(inputDF, contribs['corn'], cornCorrDF)
    inputDF['ci'] = ci['corn']
    categorizeCI(100, inputDF, title='Corn ethanol', addCount=True)

    # THIS ISNT QUITE RIGHT
    inputDF = flipCorrSign(inputDF, contribs['corn'], cornCorrDF)
    inputDF



